/* /////////////////////////////////////////////////////////////////////////
 * File:        example_cpp_3.cpp
 *
 * Purpose:     C++ example program for recls/C++. Demonstrates:
 *
 *                - stat()-ing of current directory
 *                - searching for files, according to multi-part pattern
 *                - recursive operation
 *                - evaluation of relative path of each entry, with respect
 *                  to home directory
 *                - handling exceptions and reporting of error information
 *                - elicitation of entry properties via method calls
 *
 * Created:     18th June 2006
 * Updated:     18th June 2006
 *
 * www:         http://www.recls.org/
 *
 * License:     Copyright (c) 2006, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              (Licensed under the Synesis Software Open License)
 *
 *              This source code is placed into the public domain 2006
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software.
 *
 * ////////////////////////////////////////////////////////////////////// */


/* recls Header Files */
#include <recls/cpp/fileentry.hpp>
#include <recls/cpp/filesearch.hpp>
#include <recls/implicit_link.h>

/* Standard C++ Header Files */
#include <exception>
#include <iostream>

/* Standard C Header Files */
#include <stdlib.h>     /* for EXIT_SUCCESS / EXIT_FAILURE  */

/* ////////////////////////////////////////////////////////////////////////// */

int main()
{
    try
    {
        // stat() the current directory
        recls::cpp::FileEntry   current =   recls::cpp::FileSearch::Stat("~");

        // Print out its characteristics:

        // 1. Full path
        std::cout << current.GetPath() << std::endl;

        // 2. Search-relative path
        std::cout << current.GetSearchRelativePath() << std::endl;

        // 3. Search directory
        std::cout << current.GetSearchDirectory() << std::endl;

#ifdef RECLS_PLATFORM_API_WIN32
        // 4. Drive property
        std::cout << current.GetDrive() << std::endl;
#endif /* RECLS_PLATFORM_API_WIN32 */

        // 5. Directory path property
        std::cout << current.GetDirectoryPath() << std::endl;

        // 6. Directory property
        std::cout << current.GetDirectory() << std::endl;

        // 7. File property
        std::cout << current.GetFile() << std::endl;

        // 8. File name property
        std::cout << current.GetFileName() << std::endl;

        // 9. File extension property
        std::cout << current.GetFileExt() << std::endl;

        // 10. Type
        if(current.IsDirectory())
        {
            std::cout << " <directory>" << std::endl;
        }
        else
        {
            std::cout << " <file>" << std::endl;
        }
        if(current.IsLink())
        {
            std::cout << " <link>" << std::endl;
        }
        if(current.IsReadOnly())
        {
            std::cout << " <read-only>" << std::endl;
        }

        // 11. Size
#ifdef RECLS_PLATFORM_API_WIN32
        std::cout << static_cast<unsigned long>(current.GetSize().QuadPart) << " bytes" << std::endl;
#else /* ? RECLS_PLATFORM_API_WIN32 */
        std::cout << current.GetSize() << " bytes" << std::endl;
#endif /* RECLS_PLATFORM_API_WIN32 */
    }
    catch(recls::cpp::ReclsException &x)
    {
        std::cerr << "Recls error: " << x.rc() << ", " << x.what() << std::endl;

        return EXIT_FAILURE;
    }
    catch(std::bad_alloc &)
    {
        std::cerr << "Out of memory" << std::endl;

        return EXIT_FAILURE;
    }
    catch(std::exception &x)
    {
        std::cerr << "Unhandled error: " << x.what() << std::endl;

        return EXIT_FAILURE;
    }
    catch(...)
    {
        std::cerr << "Unhandled unknown error" << std::endl;

        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

/* ////////////////////////////////////////////////////////////////////////// */
