/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/cpp/fileentry.hpp
 *
 * Purpose:     recls C++ mapping - FileEntry class.
 *
 * Created:     18th August 2003
 * Updated:     2nd June 2007
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2007, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLS_CPP_HPP_FILEENTRY
#define RECLS_INCL_RECLS_CPP_HPP_FILEENTRY

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLS_CPP_HPP_FILEENTRY_MAJOR    3
# define RECLS_VER_RECLS_CPP_HPP_FILEENTRY_MINOR    3
# define RECLS_VER_RECLS_CPP_HPP_FILEENTRY_REVISION 2
# define RECLS_VER_RECLS_CPP_HPP_FILEENTRY_EDIT     45
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

#ifdef RECLS_DOCUMENTATION_SKIP_SECTION
# define DirectoryParts_class   DirectoryParts
#endif /* RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/cpp/fileentry.hpp
 *
 * \brief [C++] Definition of the
 *  \link recls::cpp::FileEntry FileEntry\endlink
 *  and
 *  \link recls::cpp::DirectoryParts DirectoryParts\endlink
 *  classes, part of the
 *  \ref group__recls__cpp "recls C++ mapping".
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/cpp/recls.hpp>
#include <recls/assert.h>

#include <stlsoft/util/operator_bool.hpp>    // For operator bool() stuff
#if defined(RECLS_PLATFORM_IS_WIN32)
# include <winstl/filesystem/file_path_buffer.hpp>
#elif defined(RECLS_PLATFORM_IS_UNIX)
# include <unixstl/filesystem/file_path_buffer.hpp>
#else
# error Platform not (yet) recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace cpp
{
#endif /* !RECLS_NO_NAMESPACE */

#if defined(RECLS_PLATFORM_IS_WIN32)
typedef winstl::basic_file_path_buffer<recls_char_t>    file_path_buffer;
#elif defined(RECLS_PLATFORM_IS_UNIX)
typedef unixstl::basic_file_path_buffer<recls_char_t>   file_path_buffer;
#else
# error Platform not (yet) recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

class FileEntry;
class FileSearch;

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// \brief This class provides enumeration over the parts of a directory for a given file entry
///
/// \ingroup group__recls__cpp
class DirectoryParts_class
{
public:
    typedef recls_char_t            char_type;
    typedef DirectoryParts_class    class_type;

// Construction
private:
    friend class FileEntry;
    DirectoryParts_class(recls_info_t info);

/// \name Construction
/// @{
public:
    /// \brief Default constructor
    DirectoryParts_class();
    /// \brief Copy constructor
    DirectoryParts_class(class_type const &rhs);
    /// \brief Destructor
    ~DirectoryParts_class();

    /// \brief Copy assignment operator
    class_type &operator =(class_type const &rhs);

/// @}

/// \name Attributes
/// @{
public:
    /// \brief Returns the number of directory parts managed by this instance
    size_t      size() const;
    /// \brief Returns the number of directory parts managed by this instance
    ///
    /// \param index The index of the requried directory part
    ///
    /// \note The behaviour is undefined if index is out of the range denoted by \c [0, size())
    string_t    operator [](size_t index) const;

/// @}

// Implementation
private:
    static recls_info_t copy_info_(recls_info_t info);

// Members
private:
    recls_info_t    m_info;
};

/// \brief This class represents a file search entry
///
/// \ingroup group__recls__cpp
class FileEntry
{
public:
    typedef recls_char_t    char_type;
    typedef FileEntry       class_type;

// Construction
private:
    friend class Search;
    FileEntry(recls_info_t info);

/// \name Construction
/// @{
public:
    /// \brief Default constructor
    FileEntry();
    /// \brief Copy constructor
    FileEntry(FileEntry const &rhs);
    /// \brief Destructor
    ~FileEntry();

    /// \brief Copy assignment operator
    FileEntry &operator =(FileEntry const &rhs);
/// @}

/// \name State
/// @{
private:
    STLSOFT_DEFINE_OPERATOR_BOOL_TYPES(class_type, boolean_generator_type, boolean_type);
public:
    /// \brief Indicates the state of the instance, i.e. whether it is contains a valid recls_info_t or not
    operator boolean_type() const
    {
        return boolean_generator_type::translate(NULL != m_info);
    }

#ifndef STLSOFT_CF_OPERATOR_NOT_VIA_OPERATOR_POINTER_TO_MEMBER_SUPPORT
    /// \brief Returns the logical inverse of the state of the instance, i.e. whether it is contains a valid recls_info_t or not
    bool operator !() const;
#endif /* !STLSOFT_CF_OPERATOR_NOT_VIA_OPERATOR_POINTER_TO_MEMBER_SUPPORT */

/// @}

/// \name Attributes
/// @{
public:
#ifndef RECLS_PURE_API
    /// \brief Returns a pointer to the full path of the file entry, or the empty string
    char_type const     *c_str() const;

    /// \brief The length of the full path of the file entry
    size_t              length() const;
#endif /* !RECLS_PURE_API */

    /// \brief Returns the full path of the file entry
    string_t            GetPath() const;
    /// \brief Returns the directory (and drive) searched resulting in the file entry
    string_t            GetSearchDirectory() const;
    /// \brief Returns the path of the file entry relative to the search directory
    string_t            GetSearchRelativePath() const;
#ifdef RECLS_PLATFORM_API_WIN32
    /// \brief Returns the drive of the file entry
    char_type           GetDrive() const;
#endif /* RECLS_PLATFORM_API_WIN32 */
    /// \brief Returns the directory of the file entry
    string_t            GetDirectory() const;
    /// \brief Returns the directory (+ drive) of the file entry
    string_t            GetDirectoryPath() const;
    /// \brief Returns the UNC drive component, if any, of the file entry's path
    string_t            GetUNCDrive() const;
    /// \brief Returns the directory parts of the file entry
    DirectoryParts_class      GetDirectoryParts() const;
    /// \brief Returns the file (filename + extension) of the file entry
    string_t            GetFile() const;
    /// \brief Returns the short form of the file (filename + extension) of the file entry
    ///
    /// \Note On systems other than Win32, this is identical to GetFile()
    string_t            GetShortFile() const;
    /// \brief Returns the filename of the file entry
    string_t            GetFileName() const;
    /// \brief Returns the file extension of the file entry
    string_t            GetFileExt() const;

    /// \brief Returns the creation time of the file entry
    ///
    /// \Note On systems other than Win32, this is identical to GetModificationTime()
    recls_time_t        GetCreationTime() const;
    /// \brief Returns the modification time of the file entry
    recls_time_t        GetModificationTime() const;
    /// \brief Returns the last access time of the file entry
    recls_time_t        GetLastAccessTime() const;
    /// \brief Returns the last statuc change time of the file entry
    ///
    /// \Note On systems other than UNIX, this is identical to GetModificationTime()
    recls_time_t        GetLastStatusChangeTime() const;

    /// \brief Returns the size of the file entry
    recls_filesize_t    GetSize() const;

    /// \brief Indicates whether the file entry is read-only
    recls_bool_t        IsReadOnly() const;
    /// \brief Indicates whether the file entry is a directory
    recls_bool_t        IsDirectory() const;
    /// \brief Indicates whether the file entry is a link
    recls_bool_t        IsLink() const;
    /// \brief Indicates whether the file entry's path is UNC
    recls_bool_t        IsUNC() const;

    /// \brief Indicates whether the file entry exists
    recls_bool_t        Exists() const;
/// @}

/// Properties
public:
#ifdef RECLS_CPP_METHOD_PROPERTY_SUPPORT
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, Path);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, SearchDirectory);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, SearchRelativePath);
#ifdef RECLS_CPP_PLATFORM_API_WIN32
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, Drive);
#endif /* RECLS_CPP_PLATFORM_API_WIN32 */
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, Directory);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, DirectoryPath);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, UNCDrive);
#ifndef RECLS_CPP_PURE_API
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, DirectoryParts);
#endif /* !RECLS_CPP_PURE_API */
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, File);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, ShortFile);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, FileName);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, FileExt);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, CreationTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, ModificationTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, LastAccessTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, LastStatusChangeTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_DEFINE_OFFSET(class_type, Size);
union
{
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetPath, Path);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetSearchDirectory, SearchDirectory);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetSearchRelativePath, SearchRelativePath);
#ifdef RECLS_CPP_PLATFORM_API_WIN32
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(char_type, class_type, GetDrive, Drive);
#endif /* RECLS_CPP_PLATFORM_API_WIN32 */
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetDirectory, Directory);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetDirectoryPath, DirectoryPath);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetUNCDrive, UNCDrive);
#ifndef RECLS_CPP_PURE_API
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(DirectoryParts_class, class_type, GetDirectoryParts, DirectoryParts);
#endif /* !RECLS_CPP_PURE_API */
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetFile, File);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetShortFile, ShortFile);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetFileName, FileName);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(string_t, class_type, GetFileExt, FileExt);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(recls_time_t, class_type, GetCreationTime, CreationTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(recls_time_t, class_type, GetModificationTime, ModificationTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(recls_time_t, class_type, GetLastAccessTime, LastAccessTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(recls_time_t, class_type, GetLastStatusChangeTime, LastStatusChangeTime);
    RECLS_CPP_OPT_METHOD_PROPERTY_GET_PROP(recls_filesize_t, class_type, GetSize, Size);
};
#endif /* RECLS_CPP_METHOD_PROPERTY_SUPPORT */
/// @}

// Implementation
private:
    static recls_info_t copy_info_(recls_info_t info);
#ifdef RECLS_PURE_API
    // Note: this has to be inside the class because GCC is a real dummy n this regard.
    static string_t     get_string_(size_t (RECLS_CALLCONV_DEFAULT *pfn)(recls_info_t, recls_char_t *, size_t), recls_info_t info);
    {
        file_path_buffer    buffer;
        size_t              cch =   pfn(info, &buffer[0], buffer.size());

        return string_t(&buffer[0], cch);
    }
#endif /* RECLS_PURE_API */

// Members
private:
    recls_info_t    m_info;
};

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
typedef DirectoryParts_class    DirectoryParts;
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Shims
 */

#ifndef RECLS_PURE_API
/// String access shim
///
/// \ingroup group__recls__cpp
///
/// This returns a non-null null-terminated C-string of the given FileEntry's full path.
/// \param fe The FileEntry instance whose path is to be returned
///
/// \note This fits the <a href = "http://stlsoft.org/help/main.html">STLSoft</a> shims 
/// concept, described in <a href = "http://synesis.com.au/resources/articles/cpp/shims.pdf">this</a>
/// Synesis Software White Paper, and featured in the article "<a href = "http://www.cuj.com/documents/s=8681/cuj0308wilson/">Generalised String Manipulation: Access Shims and Type-tunnelling</a>",
/// in the August 2003 issue of <a href = "http://www.cuj.com">C/C++ User's Journal</a>.
inline FileEntry::char_type const *c_str_ptr(FileEntry const &fe)
{
    return fe.c_str();
}

inline FileEntry::char_type const *c_str_data(FileEntry const &fe)
{
    return fe.c_str();
}

inline size_t c_str_len(FileEntry const &fe)
{
    return fe.length();
}

# ifdef RECLS_CHAR_TYPE_IS_CHAR
inline recls_char_t const *c_str_ptr_a(FileEntry const &fe)
{
    return c_str_ptr(fe);
}

inline recls_char_t const *c_str_data_a(FileEntry const &fe)
{
    return c_str_data(fe);
}

inline size_t c_str_len_a(FileEntry const &fe)
{
    return c_str_len(fe);
}
# endif /* RECLS_CHAR_TYPE_IS_CHAR */

# ifdef RECLS_CHAR_TYPE_IS_WCHAR
inline recls_char_t const *c_str_ptr_w(FileEntry const &fe)
{
    return c_str_ptr(fe);
}

inline recls_char_t const *c_str_data_w(FileEntry const &fe)
{
    return c_str_data(fe);
}

inline size_t c_str_len_w(FileEntry const &fe)
{
    return c_str_len(fe);
}
# endif /* RECLS_CHAR_TYPE_IS_WCHAR */

#endif /* !RECLS_PURE_API */

/* /////////////////////////////////////////////////////////////////////////////
 * Implementation
 */

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION

inline /* static */ recls_info_t DirectoryParts_class::copy_info_(recls_info_t info)
{
    recls_info_t    newInfo;

    if( NULL == info ||
        RECLS_FAILED(Recls_CopyDetails(info, &newInfo)))
    {
        newInfo = NULL;
    }

    return newInfo;
}

inline DirectoryParts_class::DirectoryParts_class(recls_info_t info)
    : m_info(copy_info_(info))
{}

inline DirectoryParts_class::DirectoryParts_class()
    : m_info(NULL)
{}

inline DirectoryParts_class::DirectoryParts_class(DirectoryParts_class const &rhs)
    : m_info(copy_info_(rhs.m_info))
{}

inline DirectoryParts_class::~DirectoryParts_class()
{
    Recls_CloseDetails(m_info);
}

inline DirectoryParts_class &DirectoryParts_class::operator =(DirectoryParts_class const &rhs)
{
    Recls_CloseDetails(m_info);

    m_info = copy_info_(rhs.m_info);

    return *this;
}

inline size_t DirectoryParts_class::size() const
{
    return Recls_GetDirectoryPartProperty(m_info, -1, NULL, 0);
}

inline string_t DirectoryParts_class::operator [](size_t index) const
{
    file_path_buffer    dirPart;

    RECLS_ASSERT(NULL != m_info);
    RECLS_ASSERT(index < size());

    return Recls_GetDirectoryPartProperty(m_info, static_cast<int>(index), &dirPart[0], dirPart.size()) ? string_t(dirPart.c_str()) : string_t();
}

inline /* static */ recls_info_t FileEntry::copy_info_(recls_info_t info)
{
    recls_info_t    newInfo;

    if( NULL == info ||
        RECLS_FAILED(Recls_CopyDetails(info, &newInfo)))
    {
        newInfo = NULL;
    }

    return newInfo;
}

inline FileEntry::FileEntry(recls_info_t info)
    : m_info(info)
{}

inline FileEntry::FileEntry()
    : m_info(NULL)
{}

inline FileEntry::FileEntry(FileEntry const &rhs)
    : m_info(copy_info_(rhs.m_info))
{}

inline FileEntry::~FileEntry()
{
    if(NULL != m_info)
    {
        Recls_CloseDetails(m_info);
    }
}

inline FileEntry &FileEntry::operator =(FileEntry const &rhs)
{
    if(NULL != m_info)
    {
        Recls_CloseDetails(m_info);
    }

    m_info = copy_info_(rhs.m_info);

    return *this;
}

#ifndef STLSOFT_CF_OPERATOR_NOT_VIA_OPERATOR_POINTER_TO_MEMBER_SUPPORT
inline bool FileEntry::operator !() const
{
    return (*this) ? false : true;
}
#endif /* !STLSOFT_CF_OPERATOR_NOT_VIA_OPERATOR_POINTER_TO_MEMBER_SUPPORT */


#ifndef RECLS_PURE_API
inline FileEntry::char_type const *FileEntry::c_str() const
{
    static const char_type  s_empty[1] = { '\0' };

    return (NULL != m_info) ? m_info->path.begin : s_empty;
}

inline size_t FileEntry::length() const
{
    return (NULL != m_info) ? static_cast<size_t>(m_info->path.end - m_info->path.begin) : 0;
}
#endif /* !RECLS_PURE_API */

inline string_t FileEntry::GetPath() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetPathProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->path.begin, m_info->path.end);
#endif /* RECLS_PURE_API */
}

inline string_t FileEntry::GetSearchDirectory() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetSearchDirectoryProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->searchDirectory.begin, m_info->searchDirectory.end);
#endif /* RECLS_PURE_API */
}

inline string_t FileEntry::GetSearchRelativePath() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetSearchRelativePathProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->searchRelativePath.begin, m_info->searchRelativePath.end);
#endif /* RECLS_PURE_API */
}

#ifdef RECLS_PLATFORM_API_WIN32
inline FileEntry::char_type FileEntry::GetDrive() const
{
    char_type chDrive;

    return (Recls_GetDriveProperty(m_info, &chDrive), chDrive);
}
#endif /* RECLS_PLATFORM_API_WIN32 */

inline string_t FileEntry::GetDirectory() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetDirectoryProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->directory.begin, m_info->directory.end);
#endif /* RECLS_PURE_API */
}

inline string_t FileEntry::GetDirectoryPath() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetDirectoryPathProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->path.begin, m_info->directory.end);
#endif /* RECLS_PURE_API */
}

inline string_t FileEntry::GetUNCDrive() const
{
    RECLS_ASSERT(NULL != m_info);

#if 0
    if(!IsUNC())
    {
        return string_t();
    }
    else
    {
#ifdef RECLS_PURE_API
        string_t    directoryPath   =   GetDirectoryPath();
        size_t      directoryLen    =   Recls_GetDirectoryProperty(m_info, NULL, 0);

        return string_t(directoryPath, 0, directoryPath.length() - directoryLen);
#else /* ? RECLS_PURE_API */
        return string_t(m_info->path.begin, m_info->directory.begin);
#endif /* RECLS_PURE_API */
    }
#else /* ? 0 */
# ifdef RECLS_PURE_API
    return get_string_(Recls_GetUNCDriveProperty, m_info);
# else /* ? RECLS_PURE_API */
    return string_t(m_info->path.begin, m_info->directory.begin);
# endif /* RECLS_PURE_API */
#endif /* 0 */
}

inline recls::cpp::DirectoryParts_class FileEntry::GetDirectoryParts() const
{
    RECLS_ASSERT(NULL != m_info);

    return recls::cpp::DirectoryParts_class(m_info);
}

inline string_t FileEntry::GetFile() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetFileProperty, m_info);
#else /* ? RECLS_PURE_API */
    RECLS_ASSERT(m_info->fileName.begin <= m_info->fileExt.end);
    return string_t(m_info->fileName.begin, m_info->fileExt.end);
#endif /* RECLS_PURE_API */
}

inline string_t FileEntry::GetShortFile() const
{
#ifdef RECLS_PLATFORM_API_WIN32
    RECLS_ASSERT(NULL != m_info);

    return string_t(m_info->shortFile.begin, m_info->shortFile.end);
#else
    return GetFile();
#endif /* RECLS_PLATFORM_API_WIN32 */
}

inline string_t FileEntry::GetFileName() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetFileNameProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->fileName.begin, m_info->fileName.end);
#endif /* RECLS_PURE_API */
}

inline string_t FileEntry::GetFileExt() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetFileExtProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->fileExt.begin, m_info->fileExt.end);
#endif /* RECLS_PURE_API */
}

inline recls_time_t FileEntry::GetCreationTime() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_GetCreationTime(m_info);
}

inline recls_time_t FileEntry::GetModificationTime() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_GetModificationTime(m_info);
}

inline recls_time_t FileEntry::GetLastAccessTime() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_GetLastAccessTime(m_info);
}

inline recls_time_t FileEntry::GetLastStatusChangeTime() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_GetLastStatusChangeTime(m_info);
}

inline recls_filesize_t FileEntry::GetSize() const
{
    RECLS_ASSERT(NULL != m_info);

#ifdef RECLS_PURE_API
    recls_filesize_t    size;

    return (Recls_GetSizeProperty(m_info, &size), size);
#else /* ? RECLS_PURE_API */
    return m_info->size;
#endif /* RECLS_PURE_API */
}

inline recls_bool_t FileEntry::IsReadOnly() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_IsFileReadOnly(m_info);
}

inline recls_bool_t FileEntry::IsDirectory() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_IsFileDirectory(m_info);
}

inline recls_bool_t FileEntry::IsLink() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_IsFileLink(m_info);
}

inline recls_bool_t FileEntry::IsUNC() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_IsFileUNC(m_info);
}

inline recls_bool_t FileEntry::Exists() const
{
    RECLS_ASSERT(NULL != m_info);

    return Recls_DoesEntryExist(m_info);
}

#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace cpp */
} /* namespace recls */

#ifdef RECLS_DOCUMENTATION_SKIP_SECTION
/** \brief The recls project inserts c_str_ptr, c_str_data and c_str_len 
 * <a href = "http://www.cuj.com/documents/s=8681/cuj0308wilson/">string access shims</a> 
 * into the STLSoft namespace, for generalised manipulation of its
 * \link recls::cpp::FileEntry FileEntry\endlink type.
 */
#endif /* RECLS_DOCUMENTATION_SKIP_SECTION */
namespace stlsoft
{
//    using ::recls::cpp::c_str_ptr_null_a;
//    using ::recls::cpp::c_str_ptr_null;

# ifdef RECLS_CHAR_TYPE_IS_CHAR
    using ::recls::cpp::c_str_data_a;
    using ::recls::cpp::c_str_len_a;
    using ::recls::cpp::c_str_ptr_a;
# endif /* RECLS_CHAR_TYPE_IS_CHAR */
# ifdef RECLS_CHAR_TYPE_IS_WCHAR
    using ::recls::cpp::c_str_data_w;
    using ::recls::cpp::c_str_len_w;
    using ::recls::cpp::c_str_ptr_w;
# endif /* RECLS_CHAR_TYPE_IS_WCHAR */

    using ::recls::cpp::c_str_ptr;
    using ::recls::cpp::c_str_data;
    using ::recls::cpp::c_str_len;
}

#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef RECLS_DOCUMENTATION_SKIP_SECTION
# undef DirectoryParts_class
#endif /* RECLS_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_RECLS_CPP_HPP_FILEENTRY */

/* ////////////////////////////////////////////////////////////////////////// */
