/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_FileSearch.cpp
 *
 * Purpose:     JNI implementation file for the FileSearch class
 *
 * Created      21st November 2003
 * Updated:     16th June 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#include "recls.jni.pch.h"

#include <recls_FileSearch.h>

#include <recls/recls.h>

#include "recls.jni.util.h"

#include <stlsoft/memory/auto_buffer.hpp>
#include <stlsoft/memory/malloc_allocator.hpp>

#include <stdarg.h>

/* ////////////////////////////////////////////////////////////////////////// */

#if !defined(RECLS_NO_NAMESPACE)
using recls::hrecls_t;
using recls::recls_info_t;
using recls::recls_strptrs_t;

using recls::Recls_Search;
using recls::Recls_GetDirectoryPartProperty;
using recls::Recls_Stat;
using recls::Recls_GetRoots;
using recls::Recls_GetSelectedRoots;
using recls::Recls_CombinePaths;
using recls::Recls_DeriveRelativePath;
using recls::Recls_SqueezePath;
#endif /* !RECLS_NO_NAMESPACE */

typedef stlsoft_ns_qual(auto_buffer)<   recls_char_t
                                    ,   512
                                    ,   stlsoft_ns_qual(malloc_allocator)<recls_char_t>
                                    >       buffer_t;

/* ////////////////////////////////////////////////////////////////////////// */

/*
 * Class:     org_recls_FileSearch
 * Method:    getRoots
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_org_recls_FileSearch_getRoots
  (JNIEnv *henv, jclass cls)
{
    recls_root_t    roots[26];
    size_t          cRoots      =   Recls_GetRoots(roots, stlsoft_num_elements(roots));
    size_t          i;
    jclass          clsString   =   henv->FindClass("Ljava/lang/String;");
    jobjectArray    dirParts    =   henv->NewObjectArray(cRoots, clsString, NULL);

    for(i = 0; i < cRoots; ++i)
    {
#ifdef RECLS_CHAR_TYPE_IS_CHAR
        jstring     dirPart     =   henv->NewStringUTF(roots[i].name);
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR
        jstring     dirPart     =   henv->NewString(roots[i].name, wcslen(roots[i].name));
#else /* ?char type */
# error recls char type must be discriminated
#endif /* char type */
        henv->SetObjectArrayElement(dirParts, i, dirPart);
    }

    return dirParts;
}


/*
 * Class:     org_recls_FileSearch
 * Method:    getSelectedRoots
 * Signature: (I)[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_org_recls_FileSearch_getSelectedRoots
  (JNIEnv *henv, jclass, jint flags)
{
    recls_root_t    roots[26];
    size_t          cRoots      =   Recls_GetSelectedRoots(roots, stlsoft_num_elements(roots), flags);
    size_t          i;
    jclass          clsString   =   henv->FindClass("Ljava/lang/String;");
    jobjectArray    dirParts    =   henv->NewObjectArray(cRoots, clsString, NULL);

    for(i = 0; i < cRoots; ++i)
    {
#ifdef RECLS_CHAR_TYPE_IS_CHAR
        jstring     dirPart     =   henv->NewStringUTF(roots[i].name);
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR
        jstring     dirPart     =   henv->NewString(roots[i].name, wcslen(roots[i].name));
#else /* ?char type */
# error recls char type must be discriminated
#endif /* char type */
        henv->SetObjectArrayElement(dirParts, i, dirPart);
    }

    return dirParts;
}


/*
 * Class:     org_recls_FileSearch
 * Method:    stat
 * Signature: (Ljava/lang/String;I)Lorg/recls/Entry;
 */
JNIEXPORT jobject JNICALL Java_org_recls_FileSearch_stat
  (JNIEnv *henv, jobject obj, jstring jsPath, jint flags)
{
    recls_char_t const  *path   =   SafeStringUTFChars(henv, jsPath, NULL);
    recls_info_t        hEntry;
    recls_rc_t          rc      =   Recls_Stat(path, flags, &hEntry);
    jobject             entry;

    if(RECLS_FAILED(rc))
    {
        ThrowReclsException(henv, "Search failed", rc);

        entry = NULL;
    }
    else
    {
        jclass          clsEntry        =   henv->FindClass(CLASSNAME_ENTRY);
        jclass          clsString       =   henv->FindClass("Ljava/lang/String;");
        jmethodID       midEntryCtor    =   henv->GetMethodID(clsEntry, "<init>", "(ILjava/lang/String;[Ljava/lang/String;)V");
        size_t          cDirParts       =   Recls_GetDirectoryPartProperty(hEntry, -1, NULL, 0);
        jstring         path            =   StringFromRange(henv, hEntry->path.begin, hEntry->path.end);
        jobjectArray    dirParts        =   henv->NewObjectArray(cDirParts, clsString, NULL);

        for(size_t i = 0; i < cDirParts; ++i)
        {
            recls_strptrs_t const   *part   =   hEntry->directoryParts.begin + i;

            jstring dirPart = StringFromRange(henv, part->begin, part->end);

            henv->SetObjectArrayElement(dirParts, i, dirPart);
        }

        entry = henv->NewObject(clsEntry, midEntryCtor, hEntry, path, dirParts);
    }

    return entry;
}

/*
 * Class:     org_recls_FileSearch
 * Method:    combinePaths
 * Signature: (Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_recls_FileSearch_combinePaths
  (JNIEnv *henv, jclass, jstring jsPath1, jstring jsPath2)
{
    recls_char_t const  *path1  =   SafeStringUTFChars(henv, jsPath1, NULL);
    recls_char_t const  *path2  =   SafeStringUTFChars(henv, jsPath2, NULL);
    buffer_t            result(1 + Recls_CombinePaths(path1, path2, NULL, 0));

    if(!result.empty())
    {
        Recls_CombinePaths(path1, path2, &result[0], result.size());

        result[result.size() - 1] = '\0';
    }
    else
    {
        result[0] = '\0';
    }

#ifdef RECLS_CHAR_TYPE_IS_CHAR
    return henv->NewStringUTF(result.data());
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR
    return henv->NewString(result.data());
#else /* ?char type */
# error recls char type must be discriminated
#endif /* char type */
}

/*
 * Class:     org_recls_FileSearch
 * Method:    deriveRelativePath
 * Signature: (Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_recls_FileSearch_deriveRelativePath
  (JNIEnv *henv, jclass, jstring jsOrigin, jstring jsTarget)
{
    recls_char_t const  *origin =   SafeStringUTFChars(henv, jsOrigin, NULL);
    recls_char_t const  *target =   SafeStringUTFChars(henv, jsTarget, NULL);
    buffer_t            result(1 + Recls_DeriveRelativePath(origin, target, NULL, 0));

    if(!result.empty())
    {
        Recls_DeriveRelativePath(origin, target, &result[0], result.size());

        result[result.size() - 1] = '\0';
    }
    else
    {
        result[0] = '\0';
    }

#ifdef RECLS_CHAR_TYPE_IS_CHAR
    return henv->NewStringUTF(result.data());
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR
    return henv->NewString(result.data());
#else /* ?char type */
# error recls char type must be discriminated
#endif /* char type */
}

/*
 * Class:     org_recls_FileSearch
 * Method:    squeezePath
 * Signature: (Ljava/lang/String;I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_recls_FileSearch_squeezePath
  (JNIEnv *henv, jclass, jstring jsPath, jint width)
{
    recls_char_t const  *path   =   SafeStringUTFChars(henv, jsPath, NULL);
    buffer_t            result(1 + width);

    if(!result.empty())
    {
        Recls_SqueezePath(path, &result[0], result.size() - 1);

        result[result.size() - 1] = '\0';
    }
    else
    {
        result[0] = '\0';
    }

#ifdef RECLS_CHAR_TYPE_IS_CHAR
    return henv->NewStringUTF(result.data());
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR
    return henv->NewString(result.data());
#else /* ?char type */
# error recls char type must be discriminated
#endif /* char type */
}



/*
 * Class:     org_recls_FileSearch
 * Method:    initialise
 * Signature: (Ljava/lang/String;Ljava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_org_recls_FileSearch_initialise
  (JNIEnv *henv, jclass cls, jstring jsSearchRoot, jstring jsPattern, jint flags)
{
    // Need to 
    recls_char_t const          *searchRoot     =   SafeStringUTFChars(henv, jsSearchRoot, NULL);
    recls_char_t const          *pattern        =   SafeStringUTFChars(henv, jsPattern, NULL);

    static const recls_char_t   searchRoot_[2]  =   { '.', '\0' };
#if defined(RECLS_PLATFORM_IS_WIN32)
    static const recls_char_t   pattern_[4]     =   { '*', '.', '*', '\0' };
#elif defined(RECLS_PLATFORM_IS_UNIX)
    static const recls_char_t   pattern_[2]     =   { '*', '\0' };
#else
# error Platform not (yet) recognised
#endif /* platform */

    if(0 == *searchRoot)
    {
        searchRoot = searchRoot_;
    }
    if(0 == *pattern)
    {
        pattern = pattern_;
    }

    hrecls_t    hSrch;
    recls_rc_t  rc  =   Recls_Search(searchRoot, pattern, flags, &hSrch);

    if( RECLS_FAILED(rc) &&
        rc != RECLS_RC_NO_MORE_DATA)
    {
        ThrowReclsException(henv, "Search failed", rc);
    }

    return reinterpret_cast<jint>(hSrch);
}

/* ////////////////////////////////////////////////////////////////////////// */
