/* /////////////////////////////////////////////////////////////////////////////
 * File:        Entry.java
 *
 * Purpose:     Implementation of the org.recls.Entry class
 *
 * Created      21st November 2003
 * Updated:     16th June 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


package org.recls;

import java.util.Date;


/** This class represents an entry in the search
 */
public class Entry
{
    private Entry(int hEntry, String path, String[] directoryParts)
    {
        m_hEntry            =   hEntry;
        m_path              =   path;
        m_directoryParts    =   directoryParts;
    }
    /** Releases the resources owned by the entry
     *
     * \note It is perfectly legal to call this method more than once.
     */
    public native void close();
    /** Releases the resources owned by the entry
     */
    public void finalize()
    {
        close();
    }

// Properties
    /** The full path of the entry
     *
     * \note For "/usr/include/recls/recls_assert.h" this would yield "/usr/include/recls/recls_assert.h"
     */
    public String           getPath()
    {
        return m_path;
    }
    /** The drive component of the entry's path
     *
     * \note For "H:\Dev\include\recls\recls_assert.h" this would yield 'H'
     * \note Throws java.lang.UnsupportedOperationException if called on non-Win32 systems
     */
    public native char      getDrive(); // Throws an exception on 
    /** The directory component of the entry's path
     *
     * \note For "/usr/include/recls/recls_assert.h" this would yield "/usr/include/recls/"
     */
    public native String    getDirectory();
    /** The full location component of the entry's path.
     *
     * \note This is everything before the filename+fileext. On Win32 systems for "H:\Dev\include\recls\recls_assert.h" this would yield "H:\Dev\include\recls\"
     * \note On systems that do not have multiple drives, this returns the same value as getDirectory()
     */
    public native String    getDirectoryPath();
    /** The UNC part, if any, of the entry's path.
     *
     * \note This is only supported on Win32 systems. On other systems it will return an empty string
     */
    public native String    getUNCDrive();
    /** An array of Strings representing the parts of the Directory property
     *
     * \note For "/usr/include/recls/recls_assert.h" this would yield [ "/", "usr/", "include/", "recls/"]
     */
    public String[]         getDirectoryParts()
    {
        return m_directoryParts;
    }
    /** The file component of the entry's path
     *
     * \note For "/usr/include/recls/recls_assert.h" this would yield "recls_assert.h"
     */
    public native String    getFile();
    /** The short equivalent of the entry's File property
     *
     * \note On Win32 systems, this is the 8.3 form, e.g. "recls_~1.h". On other systems this is identical to the File property
     */
    public native String    getShortFile();
    /** The file name component of the entry's path
     *
     * \note For "/usr/include/recls/recls_assert.h" this would yield "recls_assert"
     */
    public native String    getFileName();
    /** The file extension component of the entry's path
     *
     * \note For "/usr/include/recls/recls_assert.h" this would yield "h"
     */
    public native String    getFileExt();
    /** The time the entry was created
     */
    public Date             getCreationTime()
    {
        return new Date(getCreationTime_());
    }
    /** The time the entry was last modified
     */
    public Date             getModificationTime()
    {
        return new Date(getModificationTime_());
    }
    /** The time the entry was last accessed
     */
    public Date             getLastAccessTime()
    {
        return new Date(getLastAccessTime_());
    }
    /** The time the entry's last status changed
     */
    public Date             getLastStatusChangeTime()
    {
        return new Date(getLastStatusChangeTime_());
    }
    /** The size of the entry
     */
    public native long      getSize();
    /** Indicates whether the entry is read-only
     */
    public native boolean   isReadOnly();
    /** Indicates whether the entry is a directory
     */
    public native boolean   isDirectory();
    /** Indicates whether the entry is a link
     */
    public native boolean   isLink();
    /** Indicates whether the entry is a link
     */
    public native boolean   isUNC();

    /** Provides a String representation of the entry
     */
    public String toString()
    {
        return m_path;
    }

// Implementation methods
    public native long      getCreationTime_();
    public native long      getModificationTime_();
    public native long      getLastAccessTime_();
    public native long      getLastStatusChangeTime_();

// Members
    private int         m_hEntry;   // The opaque handle to the entry
    private String      m_path;
    private String[]    m_directoryParts;
}

