/* /////////////////////////////////////////////////////////////////////////////
 * File:    recls.python.c
 *
 * Purpose: Implementation for the Python mapping for recls
 *
 * Created: 15th March 2005
 * Updated: 5th October 2006
 *
 * Home:    http://recls.org/
 *
 * Copyright (c) 2005-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the names of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Version information
 */

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_C_RECLS_PYTHON_MAJOR     1
# define RECLS_VER_C_RECLS_PYTHON_MINOR     2
# define RECLS_VER_C_RECLS_PYTHON_REVISION  1
# define RECLS_VER_C_RECLS_PYTHON_EDIT      3
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifdef _DEBUG
# undef     _DEBUG
# define    SKIP_DEBUG
#endif /* _DEBUG */

#include <Python.h>

#ifdef SKIP_DEBUG
# ifndef _DEBUG
#  define   _DEBUG
# endif /* _DEBUG */
#endif /* SKIP_DEBUG */

/* recls Header Files */
#include <recls/recls.h>
#include <recls/alloca.h>
#include <recls/assert.h>
#if defined(RECLS_PLATFORM_IS_WIN32)
# include <recls/ftp.h>
# include <winstl/time_conversion_functions.h>
#endif /* RECLS_PLATFORM_IS_WIN32 */
#include <recls/minmax.h>
#include <stlsoft/limit_traits.h>

#if !defined(RECLS_VER_1_8_1) || \
    RECLS_VER < RECLS_VER_1_8_1
# error The recls C++ mapping now requires version 1.8.1 or later of the recls core API
#endif /* RECLS_VER < 1.8.1 */

/* C Library Files */
#ifdef WIN32
# include <windows.h>
#endif /* WIn32 */

/* /////////////////////////////////////////////////////////////////////////////
 * Macros
 */

#ifndef recls_string_len_p
# define recls_string_len_p(s)  ((size_t)((s)->end - (s)->begin))
#endif /* !recls_string_len_p */
#ifndef recls_string_len_r
# define recls_string_len_r(s)  ((size_t)((s).end - (s).begin))
#endif /* !recls_string_len_r */

#ifndef NUM_ELEMENTS
# ifdef __DMC__
#  define NUM_ELEMENTS(x)       (sizeof(x) / sizeof((x)[0]))
# else /* ? compiler */
#  define NUM_ELEMENTS(x)       (sizeof(x) / sizeof(0[(x)]))
# endif /* compiler */
#endif /* !NUM_ELEMENTS */

/* ////////////////////////////////////////////////////////////////////////// */

//#define TRACE_METHODS
//#define TRACE_CTORS
//#define TRACE_DTORS

/* /////////////////////////////////////////////////////////////////////////////
 * Helper functions
 */



static PyObject *time_2_Python_time(recls_time_t t)
{
#if defined(RECLS_PURE_API) || \
    defined(RECLS_PLATFORM_IS_UNIX)
    return Py_BuildValue("l", t);
#elif defined(RECLS_PLATFORM_IS_WIN32)
    ws_sint32_t microseconds;
    time_t      seconds =   winstl__FILETIMEToUNIXTime(&t, &microseconds);
    double      d       =   (double)seconds + ((double)microseconds / 1000000.0);

    return Py_BuildValue("d", d);
#elif defined(RECLS_OVERRIDE_PLATFORM)
# error Cannot build Python with overriden platform
#else
# error Platform not (yet) recognised
#endif /* platform */
}

static PyObject *recls_filesize_2_Python_size(recls_filesize_t s)
{
#if defined(RECLS_PURE_API) || \
    defined(RECLS_PLATFORM_IS_UNIX)
    if( /* sizeof(recls_filesize_t) > sizeof(long) &&
         */s > 0x7FFFFFFF)
    {
        return PyLong_FromUnsignedLongLong(s);
    }
    else
    {
        return PyInt_FromLong((long)s);
    }
#elif defined(RECLS_PLATFORM_IS_WIN32)
    if( /* sizeof(recls_filesize_t) > sizeof(long) &&
         */s.QuadPart > 0x7FFFFFFF)
    {
        return PyLong_FromUnsignedLongLong(s.QuadPart);
    }
    else
    {
        return PyInt_FromLong((long)s.QuadPart);
    }
#elif defined(RECLS_OVERRIDE_PLATFORM)
# error Cannot build Python with overriden platform
#else
# error Platform not (yet) recognised
#endif /* platform */
}

/* /////////////////////////////////////////////////////////////////////////////
 * recls module exception type
 */

static PyObject *ReclsExceptionObject;

void recls_raise_error(recls_rc_t rc)
{
    size_t  cch =   Recls_GetErrorString(rc, NULL, 0);
    char    *s  =   (char*)recls_alloca((1 + cch) * sizeof(char));

    Recls_GetErrorString(rc, s, 1 + cch);

    PyErr_SetString(ReclsExceptionObject, s);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Entry type and method declarations
 */

staticforward PyTypeObject recls_Entry_Type;

typedef struct
{
    /* Header */
    PyObject_HEAD

    /* Entry specific information */
    recls_info_t    info;

} recls_Entry;


static recls_Entry *recls_Entry_alloc(PyObject *self, recls_info_t info);
static void         recls_Entry_dealloc(recls_Entry *self);
static int          recls_Entry_print(recls_Entry *self, FILE *file, int flags);
static PyObject     *recls_Entry_getattr(recls_Entry *self, char const *name);
static int          recls_Entry_compare(recls_Entry *lhs, recls_Entry *rhs);

static PyObject     *recls_Entry_path(recls_Entry *self);
static PyObject     *recls_Entry_searchDirectory(recls_Entry *self);
static PyObject     *recls_Entry_searchRelativePath(recls_Entry *self);
static PyObject     *recls_Entry_directory(recls_Entry *self);
static PyObject     *recls_Entry_directoryPath(recls_Entry *self);
static PyObject     *recls_Entry_uncDrive(recls_Entry *self);
static PyObject     *recls_Entry_drive(recls_Entry *self);
static PyObject     *recls_Entry_file(recls_Entry *self);
static PyObject     *recls_Entry_shortFile(recls_Entry *self);
static PyObject     *recls_Entry_fileName(recls_Entry *self);
static PyObject     *recls_Entry_fileExt(recls_Entry *self);
static PyObject     *recls_Entry_directoryParts(recls_Entry *self);
static PyObject     *recls_Entry_size(recls_Entry *self);
static PyObject     *recls_Entry_creationTime(recls_Entry *self);
static PyObject     *recls_Entry_modificationTime(recls_Entry *self);
static PyObject     *recls_Entry_lastAccessTime(recls_Entry *self);
static PyObject     *recls_Entry_lastStatusChangeTime(recls_Entry *self);

static struct PyMethodDef   recls_Entry_methods[] =
{
        {   "path",                 (PyCFunction)recls_Entry_path,                  METH_NOARGS,    "The path of the field"                                     }
    ,   {   "searchDirectory",      (PyCFunction)recls_Entry_searchDirectory,       METH_NOARGS,    "The search directory"                                      }
    ,   {   "searchRelativePath",   (PyCFunction)recls_Entry_searchRelativePath,    METH_NOARGS,    "The path of the field relative to the search directory"    }
    ,   {   "directory",            (PyCFunction)recls_Entry_directory,             METH_NOARGS,    "The directory of the field"                                }
    ,   {   "directoryPath",        (PyCFunction)recls_Entry_directoryPath,         METH_NOARGS,    "The directoryPath of the field"                            }
    ,   {   "uncDrive",             (PyCFunction)recls_Entry_uncDrive,              METH_NOARGS,    "The uncDrive of the field"                                 }
    ,   {   "drive",                (PyCFunction)recls_Entry_drive,                 METH_NOARGS,    "The drive of the field"                                    }
    ,   {   "file",                 (PyCFunction)recls_Entry_file,                  METH_NOARGS,    "The file of the field"                                     }
    ,   {   "shortFile",            (PyCFunction)recls_Entry_shortFile,             METH_NOARGS,    "The shortFile of the field"                                }
    ,   {   "fileName",             (PyCFunction)recls_Entry_fileName,              METH_NOARGS,    "The fileName of the field"                                 }
    ,   {   "fileExt",              (PyCFunction)recls_Entry_fileExt,               METH_NOARGS,    "The fileExt of the field"                                  }
    ,   {   "directoryParts",       (PyCFunction)recls_Entry_directoryParts,        METH_NOARGS,    "The directoryParts of the field"                           }
    ,   {   "size",                 (PyCFunction)recls_Entry_size,                  METH_NOARGS,    "The size of the field"                                     }
    ,   {   "creationTime",         (PyCFunction)recls_Entry_creationTime,          METH_NOARGS,    "The creationTime of the field"                             }
    ,   {   "modificationTime",     (PyCFunction)recls_Entry_modificationTime,      METH_NOARGS,    "The modificationTime of the field"                         }
    ,   {   "lastAccessTime",       (PyCFunction)recls_Entry_lastAccessTime,        METH_NOARGS,    "The lastAccessTime of the field"                           }
    ,   {   "lastStatusChangeTime", (PyCFunction)recls_Entry_lastStatusChangeTime,  METH_NOARGS,    "The lastStatusChangeTime of the field"                     }
    ,   {   NULL,                   NULL,                                           0,              NULL                                                        }
};

static PyTypeObject recls_Entry_Type =
{
        PyObject_HEAD_INIT(NULL)
        0
    ,   "Entry"
    ,   sizeof(recls_Entry)
    ,   0
    ,   (destructor)    recls_Entry_dealloc         /* tp_dealloc               */
    ,   (printfunc)     recls_Entry_print           /* tp_print                 */
    ,   (getattrfunc)   recls_Entry_getattr         /* tp_getattr               */
    ,   0                                           /* tp_setattr               */
    ,   (cmpfunc)       recls_Entry_compare         /* tp_compare               */
    ,   0                                           /* tp_repr                  */
    ,   0                                           /* tp_as_number     (PyNumberMethods*)      */
    ,   0                                           /* tp_as_sequence   (PySequenceMethods*)    */
    ,   0                                           /* tp_as_mapping    (PyMappingMethods*)     */
    ,   0                                           /* tp_hash                  */
    ,   0                                           /* tp_call                  */
    ,   0                                           /* tp_str                   */
    ,   0                                           /* tp_getattro              */
    ,   0                                           /* tp_setattro              */
    ,   0                                           /* tp_as_buffer             */
    ,   0                                           /* tp_flags                 */
    ,   0                                           /* tp_doc                   */
    ,   0                                           /* tp_traverse              */
    ,   0                                           /* tp_clear                 */
    ,   0                                           /* tp_richcompare           */
    ,   0                                           /* tp_weaklistoffset        */
    ,   0                                           /* tp_iter                  */
    ,   0                                           /* tp_iternext              */
    ,   0                                           /* tp_methods               */
    ,   0                                           /* tp_members               */
    ,   0                                           /* tp_getset                */
    ,   0                                           /* tp_base                  */
    ,   0                                           /* tp_dict                  */
    ,   0                                           /* tp_descr_get             */
    ,   0                                           /* tp_descr_set             */
    ,   0                                           /* tp_dictoffset            */
    ,   0                                           /* tp_init                  */
    ,   0                                           /* tp_alloc                 */
    ,   0                                           /* tp_new                   */
    ,   0                                           /* tp_free                  */
    ,   0                                           /* tp_is_gc                 */
    ,   0                                           /* tp_bases                 */
    ,   0                                           /* tp_mro                   */
    ,   0                                           /* tp_cache                 */
    ,   0                                           /* tp_subclasses            */
    ,   0                                           /* tp_weaklist              */
#ifdef COUNT_ALLOCS
    ,   0                                           /* tp_allocs                */
    ,   0                                           /* tp_frees                 */
    ,   0                                           /* tp_maxalloc              */
    ,   0                                           /* tp_next                  */
#endif /* COUNT_ALLOCS */
};

/* /////////////////////////////////////////////////////////////////////////////
 * FileSearch type and method declarations
 */

staticforward PyTypeObject recls_FileSearch_Type;

typedef struct
{
    /* Header */
    PyObject_HEAD

    /* FileSearch specific information */
    PyObject    *rootDir;
    PyObject    *pattern;
    long        flags;

} recls_FileSearch;


static recls_FileSearch *recls_FileSearch_alloc(PyObject *self, char const *rootDir, char const *pattern, long flags);
static void             recls_FileSearch_dealloc(recls_FileSearch *self);
static int              recls_FileSearch_print(recls_FileSearch *self, FILE *file, int flags);
static PyObject         *recls_FileSearch_getattr(recls_FileSearch *self, char const *name);

static PyObject         *recls_FileSearch_items(recls_FileSearch *self);

static struct PyMethodDef   recls_FileSearch_methods[] =
{
        {   "items",    (PyCFunction)recls_FileSearch_items,    METH_NOARGS,    "Enumerate the entries for this search" }
    ,   {   "entries",  (PyCFunction)recls_FileSearch_items,    METH_NOARGS,    "Enumerate the entries for this search; (synonym of 'items')" }
    ,   {   NULL,       NULL,                                   0,              NULL                                    }
};

static PyTypeObject recls_FileSearch_Type =
{
        PyObject_HEAD_INIT(NULL)
        0
    ,   "FileSearch"
    ,   sizeof(recls_FileSearch)
    ,   0
    ,   (destructor)    recls_FileSearch_dealloc    /* tp_dealloc   */
    ,   (printfunc)     recls_FileSearch_print      /* tp_print     */
    ,   (getattrfunc)   recls_FileSearch_getattr    /* tp_getattr   */
    ,   0                                           /* tp_setattr   */
    ,   0                                           /* tp_compare   */
};

#if defined(RECLS_PLATFORM_IS_WIN32)
/* /////////////////////////////////////////////////////////////////////////////
 * FtpSearch type and method declarations
 */

staticforward PyTypeObject recls_FtpSearch_Type;

typedef struct
{
    /* Header */
    PyObject_HEAD

    /* FtpSearch specific information */
    PyObject    *host;
    PyObject    *userName;
    PyObject    *password;
    PyObject    *rootDir;
    PyObject    *pattern;
    long        flags;

} recls_FtpSearch;


static recls_FtpSearch *recls_FtpSearch_alloc(PyObject *self, char const *host, char const *userName, char const *password, char const *rootDir, char const *pattern, long flags);
static void             recls_FtpSearch_dealloc(recls_FtpSearch *self);
static int              recls_FtpSearch_print(recls_FtpSearch *self, FILE *file, int flags);
static PyObject         *recls_FtpSearch_getattr(recls_FtpSearch *self, char const *name);

static PyObject         *recls_FtpSearch_items(recls_FtpSearch *self);

static struct PyMethodDef   recls_FtpSearch_methods[] =
{
        {   "items",    (PyCFunction)recls_FtpSearch_items,     METH_NOARGS,    "Enumerate the entries for this search" }
    ,   {   "entries",  (PyCFunction)recls_FileSearch_items,    METH_NOARGS,    "Enumerate the entries for this search; (synonym of 'items')" }
    ,   {   NULL,       NULL,                                   0,              NULL                                    }
};

static PyTypeObject recls_FtpSearch_Type =
{
        PyObject_HEAD_INIT(NULL)
        0
    ,   "FtpSearch"
    ,   sizeof(recls_FtpSearch)
    ,   0
    ,   (destructor)    recls_FtpSearch_dealloc     /* tp_dealloc   */
    ,   (printfunc)     recls_FtpSearch_print       /* tp_print     */
    ,   (getattrfunc)   recls_FtpSearch_getattr     /* tp_getattr   */
    ,   0                                           /* tp_setattr   */
    ,   0                                           /* tp_compare   */
};

#endif /* RECLS_PLATFORM_IS_WIN32 */

/* /////////////////////////////////////////////////////////////////////////////
 * SearchSequence type and method declarations
 */

staticforward PyTypeObject recls_SearchSequence_Type;

typedef struct
{
    /* Header */
    PyObject_HEAD

    /* SearchSequence specific information */
    hrecls_t        hSrch;

} recls_SearchSequence;


static recls_SearchSequence *recls_SearchSequence_alloc(PyObject *self, hrecls_t hSrch);
static void                 recls_SearchSequence_dealloc(recls_SearchSequence *self);

static PyObject             *recls_SearchSequence_item(recls_SearchSequence *self, int index);

static struct PyMethodDef   recls_SearchSequence_methods[] =
{
        {   NULL,       NULL,                   0,              NULL                            }
};

static PySequenceMethods recls_SearchSequence_as_sequence =
{
        (inquiry)           0                           /* sq_length    "len(x)"    */
    ,   (binaryfunc)        0                           /* sq_concat    "x + y"     */
    ,   (intargfunc)        0                           /* sq_repeat    "x * n"     */
    ,   (intargfunc)        recls_SearchSequence_item   /* sq_item      "x[i], in"  */
    ,   (intintargfunc)     0                           /* sq_slice     "x[i:j]"    */
    ,   (intobjargproc)     0                           /* sq_ass_item  "x[i] = v"  */
    ,   (intintobjargproc)  0                           /* sq_ass_slice "x[i:j]=v"  */
    ,   (objobjproc)        0                           /* sq_contains              */
    ,   (binaryfunc)        0                           /* sq_inplace_concat        */
    ,   (intargfunc)        0                           /* sq_inplace_repeat        */
};

static PyTypeObject recls_SearchSequence_Type =
{
        PyObject_HEAD_INIT(NULL)
        0
    ,   "SearchSequence"
    ,   sizeof(recls_SearchSequence)
    ,   0
    ,   (destructor)recls_SearchSequence_dealloc    /* tp_dealloc               */
    ,   0                                           /* tp_print                 */
    ,   0                                           /* tp_getattr               */
    ,   0                                           /* tp_setattr               */
    ,   0                                           /* tp_compare               */
    ,   0                                           /* tp_repr                  */
    ,   0                                           /* tp_as_number     (PyNumberMethods*)      */
    ,   &recls_SearchSequence_as_sequence           /* tp_as_sequence   (PySequenceMethods*)    */
    ,   0                                           /* tp_as_mapping    (PyMappingMethods*)     */
    ,   0                                           /* tp_hash                  */
    ,   0                                           /* tp_call                  */
    ,   0                                           /* tp_str                   */
    ,   0                                           /* tp_getattro              */
    ,   0                                           /* tp_setattro              */
    ,   0                                           /* tp_as_buffer             */
    ,   0                                           /* tp_flags                 */
    ,   0                                           /* tp_doc                   */
    ,   0                                           /* tp_traverse              */
    ,   0                                           /* tp_clear                 */
    ,   0                                           /* tp_richcompare           */
    ,   0                                           /* tp_weaklistoffset        */
    ,   0                                           /* tp_iter                  */
    ,   0                                           /* tp_iternext              */
    ,   0                                           /* tp_methods               */
    ,   0                                           /* tp_members               */
    ,   0                                           /* tp_getset                */
    ,   0                                           /* tp_base                  */
    ,   0                                           /* tp_dict                  */
    ,   0                                           /* tp_descr_get             */
    ,   0                                           /* tp_descr_set             */
    ,   0                                           /* tp_dictoffset            */
    ,   0                                           /* tp_init                  */
    ,   0                                           /* tp_alloc                 */
    ,   0                                           /* tp_new                   */
    ,   0                                           /* tp_free                  */
    ,   0                                           /* tp_is_gc                 */
    ,   0                                           /* tp_bases                 */
    ,   0                                           /* tp_mro                   */
    ,   0                                           /* tp_cache                 */
    ,   0                                           /* tp_subclasses            */
    ,   0                                           /* tp_weaklist              */
#ifdef COUNT_ALLOCS
    ,   0                                           /* tp_allocs                */
    ,   0                                           /* tp_frees                 */
    ,   0                                           /* tp_maxalloc              */
    ,   0                                           /* tp_next                  */
#endif /* COUNT_ALLOCS */
};

/* /////////////////////////////////////////////////////////////////////////////
 * recls module type and method declarations
 */

static PyObject *recls_stat(PyObject *self, PyObject *args);
static PyObject *recls_filesearch(PyObject *self, PyObject *args);
#if defined(RECLS_PLATFORM_IS_WIN32)
static PyObject *recls_ftpsearch(PyObject *self, PyObject *args);
#endif /* RECLS_PLATFORM_IS_WIN32 */
static PyObject *recls_roots(PyObject *self);
static PyObject *recls_selectedRoots(PyObject *self, PyObject *args);
static PyObject *recls_combinePaths(PyObject *self, PyObject *args);
static PyObject *recls_deriveRelativePath(PyObject *self, PyObject *args);
static PyObject *recls_squeezePath(PyObject *self, PyObject *args);

static struct PyMethodDef recls_methods[] =
{
        {
                "stat"
            ,   recls_stat
            ,   METH_VARARGS
            ,   "Retrieves the information for the given path, in the form of an entry object."
        }
    ,   {
                "filesearch"
            ,   recls_filesearch
            ,   METH_VARARGS
            ,   "Searches a directory for entries, according to the given pattern and (optional) flags."
        }
#if defined(RECLS_PLATFORM_IS_WIN32)
    ,   {
                "ftpsearch"
            ,   recls_ftpsearch
            ,   METH_VARARGS
            ,   "Searches an FTP host for entries, according to the given pattern and (optional) flags."
        }
#endif /* RECLS_PLATFORM_IS_WIN32 */
    ,   {
                "roots"
            ,   (PyCFunction)recls_roots
            ,   METH_NOARGS
            ,   "Returns an array of file system roots."
        }
    ,   {
                "selectedRoots"
            ,   recls_selectedRoots
            ,   METH_VARARGS
            ,   "Like 'roots', but takes a combination of flags that determine which file system roots are included."
        }
    ,   {
                "combinePaths"
            ,   recls_combinePaths
            ,   METH_VARARGS
            ,   "Combines two paths."
        }
    ,   {
                "deriveRelativePath"
            ,   recls_deriveRelativePath
            ,   METH_VARARGS
            ,   "Determines the relative path between an origin and a target path."
        }
    ,   {
                "squeezePath"
            ,   recls_squeezePath
            ,   METH_VARARGS
            ,   "Writes a path into a fixed maximum width buffer."
        }
    ,   {
                NULL
            ,   NULL
            ,   0
            ,   NULL
        }
};

/* /////////////////////////////////////////////////////////////////////////////
 * Module entry
 */

DL_EXPORT(void) initrecls(void) 
{
    PyObject    *module;

    recls_Entry_Type.ob_type = &PyType_Type;

    module  =   Py_InitModule("recls", recls_methods);

    if(NULL != module)
    {
        char        reclsVersion[101];
        char        reclsPythonVersion[101];
        PyObject    *dictionary     =   PyModule_GetDict(module);
        PyObject    *RECLS_VERSION;
        PyObject    *VERSION;
        PyObject    *FILES;
        PyObject    *DIRECTORIES;
        PyObject    *LINKS;
        PyObject    *DEVICES;
        PyObject    *TYPEMASK;
        PyObject    *RECURSIVE;
        PyObject    *NO_FOLLOW_LINKS;
        PyObject    *DIRECTORY_PARTS;
        PyObject    *DETAILS_LATER;
        PyObject    *PASSIVE_FTP;

        PyObject    *FIXED_DRIVES;
        PyObject    *NETWORK_DRIVES;
        PyObject    *CDROM_DRIVES;
        PyObject    *REMOVABLE_DRIVES;
        PyObject    *RAM_DRIVES;

        /* Create the module's exception type */
        ReclsExceptionObject    =   Py_BuildValue("s", "recls.error");

        PyDict_SetItemString(dictionary, "error", ReclsExceptionObject);

        /* Add the version constants */
        sprintf(reclsVersion, "%d.%d.%d", RECLS_VER_MAJOR, RECLS_VER_MINOR, RECLS_VER_REVISION);
        sprintf(reclsPythonVersion, "%d.%d.%d", RECLS_VER_C_RECLS_PYTHON_MAJOR, RECLS_VER_C_RECLS_PYTHON_MINOR, RECLS_VER_C_RECLS_PYTHON_REVISION);

        RECLS_VERSION       =   Py_BuildValue("s", reclsVersion);
        VERSION             =   Py_BuildValue("s", reclsPythonVersion);

        PyDict_SetItemString(dictionary, "RECLS_VERSION",RECLS_VERSION);
        PyDict_SetItemString(dictionary, "VERSION", VERSION);
        PyDict_SetItemString(dictionary, "__recls_version__",RECLS_VERSION);
        PyDict_SetItemString(dictionary, "__version__", VERSION);

        /* Add the other "constants" */
        FILES               =   PyInt_FromLong(RECLS_F_FILES);
        DIRECTORIES         =   PyInt_FromLong(RECLS_F_DIRECTORIES);
        LINKS               =   PyInt_FromLong(RECLS_F_LINKS);
        DEVICES             =   PyInt_FromLong(RECLS_F_DEVICES);
        TYPEMASK            =   PyInt_FromLong(RECLS_F_TYPEMASK);
        RECURSIVE           =   PyInt_FromLong(RECLS_F_RECURSIVE);
        NO_FOLLOW_LINKS     =   PyInt_FromLong(RECLS_F_NO_FOLLOW_LINKS);
        DIRECTORY_PARTS     =   PyInt_FromLong(RECLS_F_DIRECTORY_PARTS);
        DETAILS_LATER       =   PyInt_FromLong(RECLS_F_DETAILS_LATER);
        PASSIVE_FTP         =   PyInt_FromLong(RECLS_F_PASSIVE_FTP);

        FIXED_DRIVES        =   PyInt_FromLong(RECLS_F_FIXED_DRIVES);
        NETWORK_DRIVES      =   PyInt_FromLong(RECLS_F_NETWORK_DRIVES);
        CDROM_DRIVES        =   PyInt_FromLong(RECLS_F_CDROM_DRIVES);
        REMOVABLE_DRIVES    =   PyInt_FromLong(RECLS_F_REMOVABLE_DRIVES);
        RAM_DRIVES          =   PyInt_FromLong(RECLS_F_RAM_DRIVES);

        PyDict_SetItemString(dictionary, "FILES", FILES);
        PyDict_SetItemString(dictionary, "DIRECTORIES", DIRECTORIES);
        PyDict_SetItemString(dictionary, "LINKS", LINKS);
        PyDict_SetItemString(dictionary, "DEVICES", DEVICES);
        PyDict_SetItemString(dictionary, "TYPEMASK", TYPEMASK);
        PyDict_SetItemString(dictionary, "RECURSIVE", RECURSIVE);
        PyDict_SetItemString(dictionary, "NO_FOLLOW_LINKS", NO_FOLLOW_LINKS);
        PyDict_SetItemString(dictionary, "DIRECTORY_PARTS", DIRECTORY_PARTS);
        PyDict_SetItemString(dictionary, "DETAILS_LATER", DETAILS_LATER);
        PyDict_SetItemString(dictionary, "PASSIVE_FTP", PASSIVE_FTP);

        PyDict_SetItemString(dictionary, "FIXED_DRIVES", FIXED_DRIVES);
        PyDict_SetItemString(dictionary, "NETWORK_DRIVES", NETWORK_DRIVES);
        PyDict_SetItemString(dictionary, "CDROM_DRIVES", CDROM_DRIVES);
        PyDict_SetItemString(dictionary, "REMOVABLE_DRIVES", REMOVABLE_DRIVES);
        PyDict_SetItemString(dictionary, "RAM_DRIVES", RAM_DRIVES);
    }

    if(PyErr_Occurred())
    {
        Py_FatalError("can't initialize recls module");
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * recls module methods
 */

static PyObject *recls_stat(PyObject *self, PyObject *args)
{
    char    *path;
    long    flags   =   0;

    if(!PyArg_ParseTuple(args, "s|l", &path, &flags)) 
    {
        return NULL;
    }
    else
    {
        recls_info_t    info;
        recls_rc_t      rc;

#ifdef TRACE_METHODS
        fprintf(stderr, "recls.stat(%s, 0x%08x)\n", path, flags);
#endif /* TRACE_METHODS */

        rc  =   Recls_Stat(path, flags, &info);

        if(RECLS_FAILED(rc))
        {
            if(RECLS_RC_NO_MORE_DATA == rc)
            {
                Py_INCREF(Py_None);
                return Py_None;
            }
            else
            {
                recls_raise_error(rc);

                return NULL;
            }
        }
        else
        {
            return (PyObject*)recls_Entry_alloc(self, info);
        }
    }
}

static PyObject *recls_filesearch(PyObject *self, PyObject *args)
{
    char    *rootDir;
    char    *pattern;
    long    flags   =   0;

    if(!PyArg_ParseTuple(args, "ss|l", &rootDir, &pattern, &flags)) 
    {
        return NULL;
    }
    else
    {
#ifdef TRACE_METHODS
        fprintf(stderr, "recls.filesearch(%s, %s, 0x%08x)\n", rootDir, pattern, flags);
#endif /* TRACE_METHODS */

        return (PyObject*)recls_FileSearch_alloc(self, rootDir, pattern, flags);
    }
}

#if defined(RECLS_PLATFORM_IS_WIN32)
static PyObject *recls_ftpsearch(PyObject *self, PyObject *args)
{
    char    *host;
    char    *userName;
    char    *password;
    char    *rootDir;
    char    *pattern;
    long    flags   =   0;

    if(!PyArg_ParseTuple(args, "sssss|l", &host, &userName, &password, &rootDir, &pattern, &flags)) 
    {
        return NULL;
    }
    else
    {
#ifdef TRACE_METHODS
        fprintf(stderr, "recls.filesearch(%s, %s, %s, %s, %s, 0x%08x)\n", host, userName, "********", rootDir, pattern, flags);
#endif /* TRACE_METHODS */

        return (PyObject*)recls_FtpSearch_alloc(self, host, userName, password, rootDir, pattern, flags);
    }
}

#endif /* RECLS_PLATFORM_IS_WIN32 */

static PyObject *recls_roots(PyObject *self)
{
    size_t          numRoots    =   Recls_GetRoots(NULL, 0);
    recls_root_t    *roots      =   (recls_root_t*)recls_alloca(numRoots * sizeof(recls_root_t));
    PyObject        *tuple;

    Recls_GetRoots(roots, numRoots);

#ifdef TRACE_METHODS
    fprintf(stderr, "recls.roots(): #%u\n", numRoots);
#endif /* TRACE_METHODS */

    tuple = PyTuple_New(numRoots);

    if(NULL == tuple)
    {
        PyErr_SetFromErrno((PyObject*)self);
    }
    else
    {
        size_t  i;

        for(i = 0; i < numRoots; ++i)
        {
            PyObject    *root   =   Py_BuildValue("s", roots[i].name);

            if( NULL == root ||
                -1 == PyTuple_SetItem(tuple, i, (PyObject*)root))
            {
                Py_DECREF(tuple);

                PyErr_SetFromErrno((PyObject*)self);

                tuple = NULL;
            }
        }
    }

    return tuple;
}

static PyObject *recls_selectedRoots(PyObject *self, PyObject *args)
{
    long    flags;

    if(!PyArg_ParseTuple(args, "l", (long*)&flags)) 
    {
        return NULL;
    }
    else
    {
        size_t          numRoots    =   Recls_GetSelectedRoots(NULL, 0, (recls_uint32_t)flags);
        recls_root_t    *roots      =   (recls_root_t*)recls_alloca(numRoots * sizeof(recls_root_t));
        PyObject        *tuple;

        Recls_GetSelectedRoots(roots, numRoots, (recls_uint32_t)flags);

#ifdef TRACE_METHODS
        fprintf(stderr, "recls.selectedRoots(0x%08x): #%u\n", (recls_uint32_t)flags, numRoots);
#endif /* TRACE_METHODS */

        tuple = PyTuple_New(numRoots);

        if(NULL == tuple)
        {
            PyErr_SetFromErrno((PyObject*)self);
        }
        else
        {
            size_t  i;

            for(i = 0; i < numRoots; ++i)
            {
                PyObject    *root   =   Py_BuildValue("s", roots[i].name);

                if( NULL == root ||
                    -1 == PyTuple_SetItem(tuple, i, (PyObject*)root))
                {
                    Py_DECREF(tuple);

                    PyErr_SetFromErrno((PyObject*)self);

                    tuple = NULL;
                }
            }
        }

        return tuple;
    }
}

static PyObject *recls_combinePaths(PyObject *self, PyObject *args)
{
    char    *path1;
    char    *path2;

    if(!PyArg_ParseTuple(args, "s|s", &path1, &path2)) 
    {
        return NULL;
    }
    else
    {
        size_t          cch =   Recls_CombinePaths(path1, path2, NULL, 0);
        char *const     s   =   (char*)recls_alloca((1 + cch) * sizeof(char));

#ifdef TRACE_METHODS
        fprintf(stderr, "recls.combinePaths(%s, %s)\n", path1, path2);
#endif /* TRACE_METHODS */

        cch = Recls_CombinePaths(path1, path2, s, 1 + cch);

        return Py_BuildValue("s#", s, cch);
    }
}

static PyObject *recls_deriveRelativePath(PyObject *self, PyObject *args)
{
    char    *origin;
    char    *target;

    if(!PyArg_ParseTuple(args, "s|s", &origin, &target)) 
    {
        return NULL;
    }
    else
    {
        size_t          cch =   Recls_DeriveRelativePath(origin, target, NULL, 0);
        char *const     s   =   (char*)recls_alloca((1 + cch) * sizeof(char));

#ifdef TRACE_METHODS
        fprintf(stderr, "recls.deriveRelativePath(%s, %s)\n", origin, target);
#endif /* TRACE_METHODS */

        cch = Recls_DeriveRelativePath(origin, target, s, 1 + cch);

        return Py_BuildValue("s#", s, cch);
    }
}

static PyObject *recls_squeezePath(PyObject *self, PyObject *args)
{
    char    *path;
    size_t  width;

    if(!PyArg_ParseTuple(args, "s|l", &path, (long*)&width)) 
    {
        return NULL;
    }
    else
    {
        char *const     s   =   (char*)recls_alloca((1 + width) * sizeof(char));
        size_t          cch =   Recls_SqueezePath(path, s, width);

#ifdef TRACE_METHODS
        fprintf(stderr, "recls.squeezePath(%s, %l)\n", path, (long)width);
#endif /* TRACE_METHODS */

        return Py_BuildValue("s#", s, cch);
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * Entry methods
 */

static recls_Entry *recls_Entry_alloc(PyObject *self, recls_info_t info)
{
    recls_Entry    *f  = PyObject_New(recls_Entry, &recls_Entry_Type);

    if(NULL == f)
    {
        PyErr_SetFromErrno(self);

        Recls_CloseDetails(info);
    }
    else
    {
        f->info =   info;
    }

#ifdef TRACE_CTORS
    fprintf(stderr, "Entry::Entry(0x%08x) - %s\n", f, (NULL != f) ? "succeeded" : "failed");
#endif /* TRACE_CTORS */

    return f;
}

static void recls_Entry_dealloc(recls_Entry *self)
{
    recls_info_t    info    =   self->info;

    PyObject_Del(self);

    Recls_CloseDetails(info);

#ifdef TRACE_DTORS
    fprintf(stderr, "Entry::~Entry(0x%08x)\n", self);
#endif /* TRACE_DTORS */
}

static int recls_Entry_print(recls_Entry *self, FILE *file, int flags)
{
    if(flags & Py_PRINT_RAW)
    {
        fprintf(file, "<recls.Entry object at 0x%08x; info=0x%08x; path=%.*s>", self, self->info, (self->info->path.end - self->info->path.begin), self->info->path.begin);
    }
    else
    {
        fprintf(file, "\"<recls.Entry object at 0x%08x; info=0x%08x; path=%.*s>\"", self, self->info, (self->info->path.end - self->info->path.begin), self->info->path.begin);
    }

    return 0;
}

static PyObject *recls_Entry_getattr(recls_Entry *self, char const *name)
{
    PyMethodDef         *begin  =   &recls_Entry_methods[0];
    PyMethodDef *const  end     =   begin + (NUM_ELEMENTS(recls_Entry_methods) - 1);

#ifdef TRACE_METHODS
    fprintf(stderr, "Entry(0x%08x).__attr__(%s)\n", self, name);
#endif /* TRACE_METHODS */

    for(; begin != end; ++begin)
    {
        if(0 == strcmp(begin->ml_name, name))
        {
            return (*begin->ml_meth)((PyObject*)self, NULL);
        }
    }

    return Py_FindMethod(recls_Entry_methods, (PyObject*)self, (char*)name);
}

static int recls_Entry_compare(recls_Entry *lhs, recls_Entry *rhs)
{
    int res;

#ifdef TRACE_METHODS
    fprintf(stderr, "Entry(0x%08x).compare(0x%08x)\n", lhs, rhs);
#endif /* TRACE_METHODS */

    res = strncmp(lhs->info->path.begin, rhs->info->path.begin, recls_min((lhs->info->path.end - lhs->info->path.begin), (rhs->info->path.end - rhs->info->path.begin)));

    return res;
}

#define DEFINE_RECLS_ENTRY_STRING_ACCESSOR2(NAME, BEGIN, END)                                                       \
                                                                                                                    \
    static PyObject *recls_Entry_##NAME(recls_Entry *self)                                                          \
    {                                                                                                               \
        return Py_BuildValue("s#", self->info->BEGIN.begin, (self->info->END.end - self->info->BEGIN.begin));       \
    }

#define DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(m)                                                  \
                                                                                                \
    static PyObject *recls_Entry_##m(recls_Entry *self)                                         \
    {                                                                                           \
        return Py_BuildValue("s#", self->info->m.begin, recls_string_len_r(self->info->m));     \
    }


/* recls_Entry_path() */
static PyObject *recls_Entry_path(recls_Entry *self)
{
#ifdef TRACE_METHODS
    fprintf(stderr, "Entry(0x%08x).path()\n", self);
#endif /* TRACE_METHODS */

    return Py_BuildValue("s#", self->info->path.begin, recls_string_len_r(self->info->path));
}

/* recls_Entry_searchDirectory() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(searchDirectory)

/* recls_Entry_searchRelativePath() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(searchRelativePath)

/* recls_Entry_directory() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(directory)

/* recls_Entry_directoryPath() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR2(directoryPath, path, fileName)

/* recls_Entry_uncDrive() */
static PyObject *recls_Entry_uncDrive(recls_Entry *self)
{
    size_t      cch =   Recls_GetUNCDriveProperty(self->info, NULL, 0);
    char *const s   =   (char*)recls_alloca((1 + cch) * sizeof(char));

#ifdef TRACE_METHODS
    fprintf(stderr, "Entry(0x%08x).uncDrive()\n", self);
#endif /* TRACE_METHODS */

    cch = Recls_GetUNCDriveProperty(self->info, s, 1 + cch);

    return Py_BuildValue("s#", s, cch);
}

/* recls_Entry_drive() */
static PyObject *recls_Entry_drive(recls_Entry *self)
{
#ifdef RECLS_PLATFORM_IS_WIN32
    recls_char_t    chDrive;
#endif /* RECLS_PLATFORM_IS_WIN32 */

#ifdef TRACE_METHODS
    fprintf(stderr, "Entry(0x%08x).drive()\n", self);
#endif /* TRACE_METHODS */

#ifdef RECLS_PLATFORM_IS_WIN32
    Recls_GetDriveProperty(self->info, &chDrive);

    return Py_BuildValue("s#", &chDrive, '\0' != chDrive);
#else /* ? RECLS_PLATFORM_IS_WIN32 */
    return NULL;
#endif /* RECLS_PLATFORM_IS_WIN32 */
}

/* recls_Entry_file() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR2(file, fileName, fileExt)

/* recls_Entry_shortFile() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(shortFile)

/* recls_Entry_fileName() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(fileName)

/* recls_Entry_fileExt() */
DEFINE_RECLS_ENTRY_STRING_ACCESSOR1(fileExt)

/* recls_Entry_directoryParts() */
static PyObject *recls_Entry_directoryParts(recls_Entry *self)
{
    size_t      cParts  =   Recls_GetDirectoryPartProperty(self->info, -1, NULL, 0);

    if(0 == cParts)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }
    else
    {
        PyObject    *tuple  =   PyTuple_New(cParts);

        if(NULL == tuple)
        {
            PyErr_SetFromErrno((PyObject*)self);

            return NULL;
        }
        else
        {
            size_t  i;

            for(i = 0; i < cParts; ++i)
            {
                size_t      cch =   Recls_GetDirectoryPartProperty(self->info, i, NULL, 0);
                char        *s  =   (char*)recls_alloca((1 + cch) * sizeof(char));
                PyObject    *part;

                Recls_GetDirectoryPartProperty(self->info, i, s, 1 + cch);

                part = Py_BuildValue("s#", s, cch);

                if( NULL == part ||
                    -1 == PyTuple_SetItem(tuple, i, (PyObject*)part))
                {
                    Py_DECREF(tuple);

                    PyErr_SetFromErrno((PyObject*)self);

                    return NULL;
                }
            }
        }

        return tuple;
    }
}

/* recls_Entry_size() */
static PyObject *recls_Entry_size(recls_Entry *self)
{
    return recls_filesize_2_Python_size(self->info->size);
}

/* recls_Entry_creationTime() */
static PyObject *recls_Entry_creationTime(recls_Entry *self)
{
# if defined(RECLS_PLATFORM_IS_WIN32)
    return time_2_Python_time(self->info->creationTime);
#else
    return time_2_Python_time(self->info->modificationTime);
# endif /* RECLS_PLATFORM_IS_WIN32 */
}

/* recls_Entry_modificationTime() */
static PyObject *recls_Entry_modificationTime(recls_Entry *self)
{
    return time_2_Python_time(self->info->modificationTime);
}

/* recls_Entry_lastAccessTime() */
static PyObject *recls_Entry_lastAccessTime(recls_Entry *self)
{
    return time_2_Python_time(self->info->lastAccessTime);
}

/* recls_Entry_lastStatusChangeTime() */
static PyObject *recls_Entry_lastStatusChangeTime(recls_Entry *self)
{
# if defined(RECLS_PLATFORM_IS_UNIX)
    return time_2_Python_time(self->info->lastStatusChangeTime);
#else
    return time_2_Python_time(self->info->modificationTime);
# endif /* RECLS_PLATFORM_IS_UNIX */
}

/* /////////////////////////////////////////////////////////////////////////////
 * FileSearch methods
 */

static recls_FileSearch *recls_FileSearch_alloc(PyObject *self, char const *rootDir, char const *pattern, long flags)
{
    recls_FileSearch    *f  = PyObject_New(recls_FileSearch, &recls_FileSearch_Type);

    if(NULL == f)
    {
        PyErr_SetFromErrno(self);
    }
    else
    {
        f->rootDir  =   PyString_FromString(rootDir);
        f->pattern  =   PyString_FromString(pattern);
        f->flags    =   flags;
    }

#ifdef TRACE_CTORS
    fprintf(stderr, "FileSearch::FileSearch(0x%08x)(%s, %s, 0x%08x) - %s\n", f, rootDir, pattern, flags, (NULL != f) ? "succeeded" : "failed");
#endif /* TRACE_CTORS */

    return f;
}

static void recls_FileSearch_dealloc(recls_FileSearch *self)
{
    Py_DECREF(self->rootDir);
    Py_DECREF(self->pattern);

    PyObject_Del(self);

#ifdef TRACE_DTORS
    fprintf(stderr, "FileSearch::~FileSearch(0x%08x)\n", self);
#endif /* TRACE_DTORS */
}

static int recls_FileSearch_print(recls_FileSearch *self, FILE *file, int flags)
{
    if(flags & Py_PRINT_RAW)
    {
        fprintf(file, "<FileSearch object at 0x%08x; path=%s; pattern=%s; flags=0x%08x>", self, PyString_AsString(self->rootDir), PyString_AsString(self->pattern), self->flags);
    }
    else
    {
        fprintf(file, "\"<FileSearch object at 0x%08x; path=%s; pattern=%s; flags=0x%08x>\"", self, PyString_AsString(self->rootDir), PyString_AsString(self->pattern), self->flags);
    }

    return 0;
}

static PyObject *recls_FileSearch_getattr(recls_FileSearch *self, char const *name)
{
    if( 0 == strcmp("searchRoot", name) ||
        0 == strcmp("rootDir", name))
    {
        Py_INCREF(self->rootDir);
        return self->rootDir;
    }
    else if(0 == strcmp("pattern", name))
    {
        Py_INCREF(self->pattern);
        return self->pattern;
    }
    else if(0 == strcmp("flags", name))
    {
        return PyInt_FromLong(self->flags);
    }
    else if(0 == strcmp("len", name))
    {
        PyErr_SetString(PyExc_NotImplementedError, "recls.FileSearch.len is not available");

        return NULL;
    }

    return Py_FindMethod(recls_FileSearch_methods, (PyObject*)self, (char*)name);
}

static PyObject *recls_FileSearch_items(recls_FileSearch *self)
{
    hrecls_t    hSrch;
    recls_rc_t  rc  =   Recls_Search(PyString_AsString(self->rootDir), PyString_AsString(self->pattern), self->flags, &hSrch);

    if(RECLS_FAILED(rc))
    {
        if(RECLS_RC_NO_MORE_DATA == rc)
        {
            // Can't return NULL or throw StopIteration here

            return (PyObject*)recls_SearchSequence_alloc((PyObject*)self, NULL);
        }
        else
        {
            recls_raise_error(rc);

            return NULL;
        }
    }
    else
    {
        return (PyObject*)recls_SearchSequence_alloc((PyObject*)self, hSrch);
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * FtpSearch methods
 */

static recls_FtpSearch *recls_FtpSearch_alloc(PyObject *self, char const *host, char const *userName, char const *password, char const *rootDir, char const *pattern, long flags)
{
    recls_FtpSearch    *f  = PyObject_New(recls_FtpSearch, &recls_FtpSearch_Type);

    if(NULL == f)
    {
        PyErr_SetFromErrno(self);
    }
    else
    {
        f->host     =   PyString_FromString(host);
        f->userName =   PyString_FromString(userName);
        f->password =   PyString_FromString(password);
        f->rootDir  =   PyString_FromString(rootDir);
        f->pattern  =   PyString_FromString(pattern);
        f->flags    =   flags;
    }

#ifdef TRACE_CTORS
    fprintf(stderr, "FtpSearch::FtpSearch(0x%08x)(%s, %s, 0x%08x) - %s\n", f, rootDir, pattern, flags, (NULL != f) ? "succeeded" : "failed");
#endif /* TRACE_CTORS */

    return f;
}

static void recls_FtpSearch_dealloc(recls_FtpSearch *self)
{
    Py_DECREF(self->rootDir);
    Py_DECREF(self->pattern);

    PyObject_Del(self);

#ifdef TRACE_DTORS
    fprintf(stderr, "FtpSearch::~FtpSearch(0x%08x)\n", self);
#endif /* TRACE_DTORS */
}

static int recls_FtpSearch_print(recls_FtpSearch *self, FILE *file, int flags)
{
    if(flags & Py_PRINT_RAW)
    {
        fprintf(file, "<FtpSearch object at 0x%08x; path=%s; pattern=%s; flags=0x%08x>", self, PyString_AsString(self->rootDir), PyString_AsString(self->pattern), self->flags);
    }
    else
    {
        fprintf(file, "\"<FtpSearch object at 0x%08x; path=%s; pattern=%s; flags=0x%08x>\"", self, PyString_AsString(self->rootDir), PyString_AsString(self->pattern), self->flags);
    }

    return 0;
}

static PyObject *recls_FtpSearch_getattr(recls_FtpSearch *self, char const *name)
{
    if(0 == strcmp("host", name))
    {
        Py_INCREF(self->host);
        return self->host;
    }
    else if(0 == strcmp("userName", name))
    {
        Py_INCREF(self->userName);
        return self->userName;
    }
    else if(0 == strcmp("searchRoot", name) ||
            0 == strcmp("rootDir", name))
    {
        Py_INCREF(self->rootDir);
        return self->rootDir;
    }
    else if(0 == strcmp("pattern", name))
    {
        Py_INCREF(self->pattern);
        return self->pattern;
    }
    else if(0 == strcmp("flags", name))
    {
        return PyInt_FromLong(self->flags);
    }
    else if(0 == strcmp("len", name))
    {
        PyErr_SetString(PyExc_NotImplementedError, "recls.FtpSearch.len is not available");

        return NULL;
    }

    return Py_FindMethod(recls_FtpSearch_methods, (PyObject*)self, (char*)name);
}

static PyObject *recls_FtpSearch_items(recls_FtpSearch *self)
{
    hrecls_t    hSrch;
    recls_rc_t  rc  =   Recls_SearchFtp(PyString_AsString(self->host), PyString_AsString(self->userName), PyString_AsString(self->password), PyString_AsString(self->rootDir), PyString_AsString(self->pattern), self->flags, &hSrch);

    if(RECLS_FAILED(rc))
    {
        if(RECLS_RC_NO_MORE_DATA == rc)
        {
            // Can't return NULL or throw StopIteration here

            return (PyObject*)recls_SearchSequence_alloc((PyObject*)self, NULL);
        }
        else
        {
            recls_raise_error(rc);

            return NULL;
        }
    }
    else
    {
        return (PyObject*)recls_SearchSequence_alloc((PyObject*)self, hSrch);
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * SearchSequence methods
 */

static recls_SearchSequence *recls_SearchSequence_alloc(PyObject *self, hrecls_t hSrch)
{
    recls_SearchSequence    *f  = PyObject_New(recls_SearchSequence, &recls_SearchSequence_Type);

    if(NULL == f)
    {
        PyErr_SetFromErrno(self);

        Recls_SearchClose(hSrch);
    }
    else
    {
        f->hSrch    =   hSrch;
    }

#ifdef TRACE_CTORS
    fprintf(stderr, "SearchSequence::SearchSequence(0x%08x) - %s\n", f, (NULL != f) ? "succeeded" : "failed");
#endif /* TRACE_CTORS */

    return f;
}

static void recls_SearchSequence_dealloc(recls_SearchSequence *self)
{
    hrecls_t    hSrch   =   self->hSrch;

    PyObject_Del(self);

    if(NULL != hSrch)
    {
        Recls_SearchClose(hSrch);
    }

#ifdef TRACE_DTORS
    fprintf(stderr, "SearchSequence::~SearchSequence(0x%08x)\n", self);
#endif /* TRACE_DTORS */
}

static PyObject *recls_SearchSequence_item(recls_SearchSequence *self, int index)
{
    recls_info_t    info;
    recls_rc_t      rc;

#ifdef TRACE_METHODS
    fprintf(stderr, "SearchSequence(0x%08x).__item__(%d)\n", self, index);
#endif /* TRACE_METHODS */

    if(NULL == self->hSrch)
    {
        rc = RECLS_RC_NO_MORE_DATA;
    }
    else
    {
        if(0 == index)
        {
            rc = Recls_GetDetails(self->hSrch, &info);
        }
        else
        {
            rc = Recls_GetNextDetails(self->hSrch, &info);
        }
    }

    if(RECLS_FAILED(rc))
    {
        if(RECLS_RC_NO_MORE_DATA == rc)
        {
            PyErr_SetString(PyExc_StopIteration, "");

            return NULL;
        }
        else
        {
            recls_raise_error(rc);

            return NULL;
        }
    }
    else
    {
        return (PyObject*)recls_Entry_alloc((PyObject*)self, info);
    }
}

/* ////////////////////////////////////////////////////////////////////////// */
