/* /////////////////////////////////////////////////////////////////////////////
 * File:        entryinfo.cpp (formerly part of ReclsFileSearchDirectoryNode.cpp)
 *
 * Purpose:     Implementation of the create_entryinfo() function.
 *
 * Created:     31st May 2004
 * Updated:     27th February 2007
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2004-2007, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/recls.h>
#include <recls/assert.h>
#include "recls_string.h"
#include "recls_util.h"
#include "recls_impl.h"

#if defined(RECLS_CHAR_TYPE_IS_WCHAR)
# if defined(RECLS_PLATFORM_IS_WIN32)
#  include <winstl/conversion/char_conversions.hpp>
# endif /* RECLS_PLATFORM_IS_WIN32 */
#endif /* RECLS_CHAR_TYPE_IS_???? */

#include "recls_debug.h"

#include "ReclsFileSearchDirectoryNode.hpp"
#include "entryinfo.hpp"
#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

typedef ReclsFileSearchDirectoryNode::traits_type               traits_type;
typedef ReclsFileSearchDirectoryNode::file_path_buffer_type     file_path_buffer_type;

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * Utility functions
 */

static size_t count_dir_parts(recls_char_t const *begin, recls_char_t const *end)
{
    return count_char_instances(begin, end, traits_type::path_name_separator());
}

recls_info_t create_entryinfo(  size_t                              rootDirLen
                            ,   recls_char_t const                  *searchDir
                            ,   size_t                              searchDirLen
                            ,   recls_char_t const                  *entryPath
                            ,   size_t                              entryPathLen
                            ,   recls_char_t const                  *entryFile
                            ,   size_t                              entryFileLen
                            ,   recls_uint32_t                      flags
                            ,   traits_type::stat_data_type const   *st)
{
    function_scope_trace("create_entryinfo");

	const int		bSearchDirOverlap	=	0 == traits_type::str_n_compare(searchDir, entryPath, searchDirLen);

    STLSOFT_SUPPRESS_UNUSED(searchDir);
    STLSOFT_SUPPRESS_UNUSED(entryFile);

    RECLS_MESSAGE_ASSERT("Directory length cannot be 0", 0 != searchDirLen);
    RECLS_MESSAGE_ASSERT("Path length cannot be 0", 0 != entryPathLen);
    RECLS_MESSAGE_ASSERT("File length cannot be 0", 0 != entryFileLen);

    RECLS_MESSAGE_ASSERT("Invalid directory length", searchDirLen <= traits_type::str_len(searchDir));
    RECLS_MESSAGE_ASSERT("Invalid path length", entryPathLen == traits_type::str_len(entryPath));
    RECLS_MESSAGE_ASSERT("Invalid file length", entryFileLen == traits_type::str_len(entryFile));

    RECLS_ASSERT(!bSearchDirOverlap || searchDirLen < entryPathLen);
    RECLS_ASSERT(!bSearchDirOverlap || searchDirLen + entryFileLen <= entryPathLen);

    RECLS_ASSERT(!bSearchDirOverlap || 0 == traits_type::str_n_compare(searchDir, entryPath, searchDirLen));
    RECLS_ASSERT(0 == traits_type::str_n_compare(entryFile, entryPath + (entryPathLen - entryFileLen), entryFileLen));

    RECLS_ASSERT(searchDir[searchDirLen - 1] == traits_type::path_name_separator());

    RECLS_ASSERT(traits_type::str_rchr(entryPath, traits_type::path_name_separator()) + 1 == entryPath + (entryPathLen - entryFileLen));

    // size of structure is:
    //
    //    offsetof(struct recls_fileinfo_t, data)
    //  + directory parts
    //  + full path (+ null)
    //  + short name (+ null)
    //  + 1 in case we need to MARK_DIRS
	//  [+ 1 + searchDirLen if searchDir and entryPath do not overlap]

    recls_char_t const *const   dir0    =   find_directory_0(entryPath);
    recls_char_t const *const   end     =   entryPath + (entryPathLen - entryFileLen);

    const size_t    cchFileName =   entryFileLen;
    const size_t    cDirParts   =   (RECLS_F_DIRECTORY_PARTS == (flags & RECLS_F_DIRECTORY_PARTS)) ? count_dir_parts(dir0, end) : 0;
    const size_t    cbPath      =   align_up_size(sizeof(recls_char_t) * (1 + entryPathLen));
#if defined(RECLS_PLATFORM_IS_UNIX)
    const size_t    cbAlt       =   0;  // UNIX doesn't have alt paths
#elif defined(RECLS_PLATFORM_IS_WIN32)
    const size_t    cbAlt       =   align_up_size(sizeof(recls_char_t) * (1 + RECLS_NUM_ELEMENTS(st->cAlternateFileName)));
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */
    const size_t    cb          =   offsetof(struct recls_fileinfo_t, data)
                                +   cDirParts * sizeof(recls_strptrs_t)
                                +   cbPath
                                +   cbAlt
                                +   1 // In case we need to expand for MARK_DIRS
								+	(bSearchDirOverlap ? 0 : (1 + searchDirLen + 1))
                                ;

    struct recls_fileinfo_t *info   =   const_cast<struct recls_fileinfo_t*>(Entry_Allocate(cb));

    if(NULL != info)
    {
        recls_byte_t* const     pData       =   &info->data[0];
        recls_byte_t* const     pParts      =   pData + 0;
        recls_byte_t* const     pPath       =   pParts + (cDirParts * sizeof(recls_strptrs_t));
#if defined(RECLS_PLATFORM_IS_WIN32)
        recls_byte_t* const     pAltName    =   pPath + cbPath;
#endif /* platform */
		recls_byte_t* const		pSearchCopy	=	pPath + cbPath + cbAlt;

        recls_char_t            *fullPath   =   ::stlsoft::sap_cast<recls_char_t*>(pPath);
#if defined(RECLS_PLATFORM_IS_WIN32)
        recls_char_t            *altName    =   ::stlsoft::sap_cast<recls_char_t*>(pAltName);
#endif /* platform */

        RECLS_ASSERT(::stlsoft::sap_cast<recls_char_t*>(pData + (cDirParts * sizeof(recls_strptrs_t))) == fullPath);
#if defined(RECLS_PLATFORM_IS_WIN32)
        RECLS_ASSERT(::stlsoft::sap_cast<recls_char_t*>(pData + (cDirParts * sizeof(recls_strptrs_t) + cbPath)) == altName);
#endif /* platform */

        info->checkSum                      =   0;
        info->extendedFlags[0]              =   0;
        info->extendedFlags[1]              =   0;

        // full path
        traits_type::str_n_copy(&fullPath[0], entryPath, entryPathLen + 1);
        info->path.begin                    =   fullPath;
        info->path.end                      =   fullPath + entryPathLen;

        // search-relative path
		if(bSearchDirOverlap)
		{
	        info->searchRelativePath.begin      =   info->path.begin + rootDirLen;
		    info->searchRelativePath.end        =   info->path.end;
		}
		else
		{
	        info->searchRelativePath.begin      =   info->path.begin;
		    info->searchRelativePath.end        =   info->path.end;
		}

        // drive, directory, file (name + ext)
#if defined(RECLS_PLATFORM_IS_UNIX)
        info->directory.begin               =   &fullPath[dir0 - entryPath];
#elif defined(RECLS_PLATFORM_IS_WIN32)
        info->drive                         =   ('\\' == fullPath[0]) ? '\0' : fullPath[0];
        info->directory.begin               =   &fullPath[dir0 - entryPath];
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */
        info->directory.end                 =   fullPath + (entryPathLen - entryFileLen);
        info->fileName.begin                =   info->directory.end;
        info->fileName.end                  =   traits_type::str_rchr(info->directory.end, RECLS_LITERAL('.'));
        if(NULL != info->fileName.end)
        {
            info->fileExt.begin             =   info->fileName.end + 1;
            info->fileExt.end               =   info->directory.end +  cchFileName;
        }
        else
        {
            info->fileName.end              =   info->directory.end +  cchFileName;
            info->fileExt.begin             =   info->directory.end +  cchFileName;
            info->fileExt.end               =   info->directory.end +  cchFileName;
        }

        // determine the directory parts
        recls_char_t const      *p          =   info->directory.begin;
        recls_char_t const      *l          =   info->directory.end;
        struct recls_strptrs_t  *begin      =   ::stlsoft::sap_cast<struct recls_strptrs_t*>(&info->data[0]);

        info->directoryParts.begin          =   begin;
        info->directoryParts.end            =   begin + cDirParts;

		if(bSearchDirOverlap)
		{
	        info->searchDirectory.begin     =   info->path.begin;
		    info->searchDirectory.end       =   info->searchDirectory.begin + rootDirLen;
		}
		else
		{
			recls_char_t	*searchCopy		=	::stlsoft::sap_cast<recls_char_t*>(pSearchCopy);
	        info->searchDirectory.begin     =   searchCopy;
		    info->searchDirectory.end       =   info->searchDirectory.begin + searchDirLen;
			traits_type::str_n_copy(searchCopy, searchDir, 1 + searchDirLen);
		}

        if(info->directoryParts.begin != info->directoryParts.end)
        {
            RECLS_ASSERT((flags & RECLS_F_DIRECTORY_PARTS) == RECLS_F_DIRECTORY_PARTS);

            begin->begin = p;

            for(; p != l; ++p)
            {
                if(*p == traits_type::path_name_separator())
                {
                    begin->end = p + 1;

                    if(++begin != info->directoryParts.end)
                    {
                        begin->begin = p + 1;
                    }
                }
            }
        }

        if(NULL == st)
        {
            recls_time_t        no_time;
            recls_filesize_t    no_size;

            memset(&no_time, 0, sizeof(no_time));
            memset(&no_size, 0, sizeof(no_size));

            // alt name
#if defined(RECLS_PLATFORM_IS_WIN32)
            altName[0] = '\0';
            info->shortFile.begin   =   altName;
            info->shortFile.end     =   altName;
#endif /* platform */

            // attributes
#if defined(RECLS_PLATFORM_IS_UNIX)
            info->lastStatusChangeTime  =   no_time;
#elif defined(RECLS_PLATFORM_IS_WIN32)
            info->creationTime          =   no_time;
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */
            info->modificationTime      =   no_time;
            info->lastAccessTime        =   no_time;
            info->size                  =   no_size;
        }
        else
        {
            // alt name
#if defined(RECLS_PLATFORM_IS_WIN32)
            traits_type::str_copy(altName, st->cAlternateFileName);
            info->shortFile.begin   =   altName;
            info->shortFile.end     =   altName + traits_type::str_len(altName);
#endif /* platform */

            // attributes
#if defined(RECLS_PLATFORM_IS_UNIX)
            info->attributes            =   st->st_mode;
#elif defined(RECLS_PLATFORM_IS_WIN32)
            info->attributes            =   st->dwFileAttributes;
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */

            // time, size
#if defined(RECLS_PLATFORM_IS_UNIX)
            info->lastStatusChangeTime  =   st->st_ctime;
            info->modificationTime      =   st->st_mtime;
            info->lastAccessTime        =   st->st_atime;
            info->size                  =   st->st_size;
#elif defined(RECLS_PLATFORM_IS_WIN32)
            info->creationTime          =   st->ftCreationTime;
            info->modificationTime      =   st->ftLastWriteTime;
            info->lastAccessTime        =   st->ftLastAccessTime;
            info->size.u.HighPart       =   st->nFileSizeHigh;
            info->size.u.LowPart        =   st->nFileSizeLow;
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */

            // Handle MARK_DIRS
            if( RECLS_F_MARK_DIRS == (flags & RECLS_F_MARK_DIRS) &&
                traits_type::is_directory(st))
            {
                *const_cast<recls_char_t*>(info->path.end)  =   traits_type::path_name_separator();
                ++info->path.end;
                *const_cast<recls_char_t*>(info->path.end)  =   '\0';
            }
        }

        // Checks
        RECLS_ASSERT(info->path.begin < info->path.end);
        RECLS_ASSERT(info->directory.begin < info->directory.end);
        RECLS_ASSERT(info->path.begin <= info->directory.begin);
        RECLS_ASSERT(info->directory.end <= info->path.end);
        RECLS_ASSERT(info->fileName.begin <= info->fileName.end);
        RECLS_ASSERT(info->fileExt.begin <= info->fileExt.end);
        RECLS_ASSERT(info->fileName.begin < info->fileExt.end);
        RECLS_ASSERT(info->fileName.end <= info->fileExt.begin);
    }

    return info;
}


recls_info_t create_drive_entryinfo(recls_char_t const                                              *entryPath
                                ,   size_t                                                          entryPathLen
                                ,   recls_uint32_t                                                  flags
                                ,   ReclsFileSearchDirectoryNode::traits_type::stat_data_type const *st)
{
#if 0
    return create_entryinfo(0, NULL, 0, entryPath, entryPathLen, NULL, 0, flags, st);
#else /* ? 0 */
    const size_t    cDirParts   =   0;
    const size_t    cbPath      =   align_up_size(sizeof(recls_char_t) * (1 + entryPathLen));
    const size_t    cb          =   offsetof(struct recls_fileinfo_t, data)
                                +   cbPath
                                +   1 // In case we need to expand for MARK_DIRS
                                ;

    struct recls_fileinfo_t *info   =   const_cast<struct recls_fileinfo_t*>(Entry_Allocate(cb));

    if(NULL != info)
    {
        recls_byte_t* const     pData       =   &info->data[0];
        recls_byte_t* const     pParts      =   pData + 0;
        recls_byte_t* const     pPath       =   pParts + (cDirParts * sizeof(recls_strptrs_t));
        recls_char_t            *fullPath   =   ::stlsoft::sap_cast<recls_char_t*>(pPath);

        RECLS_ASSERT(::stlsoft::sap_cast<recls_char_t*>(pData + (cDirParts * sizeof(recls_strptrs_t))) == fullPath);

        info->checkSum                      =   0;
        info->extendedFlags[0]              =   0;
        info->extendedFlags[1]              =   0;

        // full path
        traits_type::str_n_copy(&fullPath[0], entryPath, entryPathLen + 1);
        info->path.begin                    =   fullPath;
        info->path.end                      =   fullPath + entryPathLen;

        // search-relative path
        info->searchRelativePath.begin      =   info->path.begin;
        info->searchRelativePath.end        =   info->path.end;

        // drive, directory, file (name + ext)
        info->directory.begin               =   info->path.end;
#if defined(RECLS_PLATFORM_IS_WIN32)
        info->drive                         =   ('\\' == fullPath[0]) ? '\0' : fullPath[0];
#elif defined(RECLS_PLATFORM_IS_UNIX)
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */
        info->directory.begin               =   info->path.end;
        info->directory.end                 =   info->path.end;
        info->fileName.begin                =   info->path.end;
        info->fileName.end                  =   info->path.end;
        info->fileExt.begin                 =   info->path.end;
        info->fileExt.end                   =   info->path.end;

        // determine the directory parts
        info->directoryParts.begin          =   NULL;
        info->directoryParts.end            =   NULL;

        info->searchDirectory.begin         =   info->path.end;
        info->searchDirectory.end           =   info->path.end;

        if(NULL == st)
        {
            recls_time_t        no_time;
            recls_filesize_t    no_size;

            memset(&no_time, 0, sizeof(no_time));
            memset(&no_size, 0, sizeof(no_size));

            // alt name
#if defined(RECLS_PLATFORM_IS_WIN32)
            info->shortFile.begin           =   info->path.end;
            info->shortFile.end             =   info->path.end;
#endif /* platform */

            // attributes
#if defined(RECLS_PLATFORM_IS_UNIX)
            info->lastStatusChangeTime  =   no_time;
#elif defined(RECLS_PLATFORM_IS_WIN32)
            info->creationTime          =   no_time;
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */
            info->modificationTime      =   no_time;
            info->lastAccessTime        =   no_time;
            info->size                  =   no_size;
        }
        else
        {
            // alt name
#if defined(RECLS_PLATFORM_IS_WIN32)
            info->shortFile.begin           =   info->path.begin;
            info->shortFile.end             =   info->path.end;
#endif /* platform */

            // attributes
#if defined(RECLS_PLATFORM_IS_UNIX)
            info->attributes            =   st->st_mode;
#elif defined(RECLS_PLATFORM_IS_WIN32)
            info->attributes            =   st->dwFileAttributes;
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */

            // time, size
#if defined(RECLS_PLATFORM_IS_UNIX)
            info->lastStatusChangeTime  =   st->st_ctime;
            info->modificationTime      =   st->st_mtime;
            info->lastAccessTime        =   st->st_atime;
            info->size                  =   st->st_size;
#elif defined(RECLS_PLATFORM_IS_WIN32)
            info->creationTime          =   st->ftCreationTime;
            info->modificationTime      =   st->ftLastWriteTime;
            info->lastAccessTime        =   st->ftLastAccessTime;
            info->size.u.HighPart       =   st->nFileSizeHigh;
            info->size.u.LowPart        =   st->nFileSizeLow;
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */

            // Handle MARK_DIRS
            if( RECLS_F_MARK_DIRS == (flags & RECLS_F_MARK_DIRS) &&
                traits_type::is_directory(st))
            {
                *const_cast<recls_char_t*>(info->path.end)  =   traits_type::path_name_separator();
                ++info->path.end;
                *const_cast<recls_char_t*>(info->path.end)  =   '\0';
            }
        }

        // Checks
        RECLS_ASSERT(info->path.begin < info->path.end);
        RECLS_ASSERT(info->directory.begin == info->directory.end);
        RECLS_ASSERT(info->path.begin < info->directory.begin);
        RECLS_ASSERT(info->directory.end == info->path.end);
        RECLS_ASSERT(info->fileName.begin == info->fileName.end);
        RECLS_ASSERT(info->fileExt.begin == info->fileExt.end);
        RECLS_ASSERT(info->fileName.begin == info->fileExt.end);
        RECLS_ASSERT(info->fileName.end == info->fileExt.begin);
    }

    return info;
#endif /* 0 */
}

/* /////////////////////////////////////////////////////////////////////////
 * Unit-tests
 */

RECLS_FNDECL(void) Recls_RunUnittests_EntryInfo(FILE *stm)
{
	STLSOFT_SUPPRESS_UNUSED(stm);

#if 0
	recls_info_t	info;

	// 1.
	info = create_entryinfo(   size_t                              rootDirLen
                            ,   recls_char_t const                  *searchDir
                            ,   size_t                              searchDirLen
                            ,   recls_char_t const                  *entryPath
                            ,   size_t                              entryPathLen
                            ,   recls_char_t const                  *entryFile
                            ,   size_t                              entryFileLen
                            ,   recls_uint32_t                      flags
                            ,   traits_type::stat_data_type const   *st)
#endif /* 0 */

	;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
