/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_api.cpp
 *
 * Purpose:     Main (platform-independent) implementation file for the recls API.
 *
 * Created:     16th August 2003
 * Updated:     2nd June 2007
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2007, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/recls.h>
#include <recls/assert.h>
#include "recls_impl.h"
#include "recls_util.h"     // for recls_strlen(), recls_strncpy()

#include <stlsoft/stlsoft.h>    /* If the compiler cannot find this, you are not using STLSoft 1.9.1 or later, as required. */
#if !defined(_STLSOFT_VER_1_9_1) || \
    _STLSOFT_VER < _STLSOFT_VER_1_9_1
# error Requires STLSoft 1.9.1, or later. (www.stlsoft.org/downloads.html)
#endif /* STLSoft version */

#include <stlsoft/memory/auto_buffer.hpp>
//#include <stlsoft/shims/access/string.hpp>

# include <platformstl/filesystem/filesystem_traits.hpp>
# include <platformstl/filesystem/file_path_buffer.hpp>
# include <platformstl/filesystem/path.hpp>
# include <platformstl/filesystem/path_functions.hpp>


#include "recls_debug.h"

#include "ReclsSearch.hpp"
#include "ReclsFileSearch.hpp"
#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

typedef ::platformstl::filesystem_traits<recls_char_t>      traits_type;
typedef ::platformstl::basic_file_path_buffer<recls_char_t> file_path_buffer_type;
typedef ::platformstl::basic_path<recls_char_t>             path_type;

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

#if defined(RECLS_PLATFORM_IS_WIN32x)
struct cdecl_process_fn_translator
{
    typedef int (RECLS_CALLCONV_STDDECL *stdcall_process_fn_t)( recls_info_t                hEntry
                                                            ,   recls_process_fn_param_t    param);

    cdecl_process_fn_translator(hrecls_process_fn_t         pfn
                            ,   recls_process_fn_param_t    param)
        : m_pfn((stdcall_process_fn_t)pfn)
        , m_param(param)
    {}

    operator recls_process_fn_param_t()
    {
        return this;
    }

    static int RECLS_CALLCONV_DEFAULT func( recls_info_t                hEntry
                                        ,   recls_process_fn_param_t    param)
    {
        cdecl_process_fn_translator *pThis  =   (cdecl_process_fn_translator*)param;

        return (*pThis->m_pfn)(hEntry, pThis->m_param);
    }

private:
    stdcall_process_fn_t        m_pfn;
    recls_process_fn_param_t    m_param;
};

struct cdecl_progress_fn_translator
{
    typedef int (RECLS_CALLCONV_STDDECL *stdcall_progress_fn_t)(recls_char_t const          *dir
                                                            ,   size_t                      dirLen
                                                            ,   recls_process_fn_param_t    param
                                                            ,   void                        *reserved0
                                                            ,   recls_uint32_t              reserved1);

    cdecl_progress_fn_translator(hrecls_progress_fn_t           pfn
                            ,   recls_process_fn_param_t    param)
        : m_pfn((stdcall_progress_fn_t)pfn)
        , m_param(param)
    {}

    operator recls_process_fn_param_t()
    {
        return this;
    }

    static int RECLS_CALLCONV_DEFAULT func( recls_char_t const          *dir
                                        ,   size_t                      dirLen
                                        ,   recls_process_fn_param_t    param
                                        ,   void                        *reserved0
                                        ,   recls_uint32_t              reserved1)
    {
        cdecl_progress_fn_translator    *pThis  =   (cdecl_progress_fn_translator*)param;

        return (*pThis->m_pfn)(dir, dirLen, pThis->m_param, reserved0, reserved1);
    }

private:
    stdcall_progress_fn_t       m_pfn;
    recls_process_fn_param_t    m_param;
};

#endif /* RECLS_PLATFORM_IS_WIN32 */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

static const recls_char_t   s_defaultSearchRoot[]       =   RECLS_LITERAL(".");
static const recls_char_t   s_home[]                    =   RECLS_LITERAL("~");

#if defined(RECLS_PLATFORM_IS_UNIX)
# define _GetCreationTime           modificationTime
# define _GetLastStatusChangeTime   lastStatusChangeTime
# define RAPI_PATHNAMESEP           '/'
# define RAPI_PATHSEP               ':'
# define RAPI_WILDCARDSALL          '*'
#elif defined(RECLS_PLATFORM_IS_WIN32)
# define _GetCreationTime           creationTime
# define _GetLastStatusChangeTime   modificationTime
# define RAPI_PATHNAMESEP           '\\'
# define RAPI_PATHSEP               ';'
# define RAPI_WILDCARDSALL          '*', '.', '*'
#else /* unrecognised platform */
# error platform is not recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Search control
 */

RECLS_FNDECL(recls_rc_t) Recls_Search(  recls_char_t const  *searchRoot
                                    ,   recls_char_t const  *pattern
                                    ,   recls_uint32_t      flags
                                    ,   hrecls_t            *phSrch)
{
    function_scope_trace("Recls_Search");

    return Recls_SearchFeedback(searchRoot, pattern, flags, NULL, NULL, phSrch);
}

RECLS_FNDECL(recls_rc_t) Recls_SearchFeedback(  recls_char_t const          *searchRoot
                                    ,           recls_char_t const          *pattern
                                    ,           recls_uint32_t              flags
                                    ,           hrecls_progress_fn_t        pfn
                                    ,           recls_process_fn_param_t    param
                                    ,           hrecls_t                    *phSrch)
{
    function_scope_trace("Recls_SearchFeedback");
    debug_printf("Recls_SearchFeedback(%s, %s, 0x%04x, . . .)\n", searchRoot, pattern, flags);

#if defined(RECLS_PLATFORM_IS_WIN32x)
    cdecl_progress_fn_translator    translator(pfn, param);

    if( NULL != pfn &&
        (flags & RECLS_F_CALLBACKS_STDCALL_ON_WIN32))
    {
        pfn     =   &cdecl_progress_fn_translator::func;
        param   =   translator;
    }
#endif /* RECLS_PLATFORM_IS_WIN32 */

    RECLS_ASSERT(NULL != phSrch);

    *phSrch = static_cast<hrecls_t>(0);

    recls_rc_t  rc;

#ifdef _RECLS_EXCEPTION_SUPPORT
    try
    {
#endif /* _RECLS_EXCEPTION_SUPPORT */
        // Default the input parameters
        file_path_buffer_type   home;

        // Default the pattern?
        if(NULL == pattern)
        {
            // Pre-1.8.6, this always defaulted to '*.*' (or '*')

            if( NULL != searchRoot &&
                (   NULL != traits_type::str_pbrk(searchRoot, RECLS_LITERAL("*?")) ||
                    traits_type::is_file(searchRoot)))
            {
                std::swap(searchRoot, pattern);
            }
            else
            {
                // Set to '*' / '*.*'
                pattern = traits_type::pattern_all();
            }
        }

        // Default the search root?
        if( NULL == searchRoot ||
            0 == *searchRoot)
        {
            if(USE_TILDE_ON_NO_SEARCHROOT & flags)
            {
                searchRoot = s_home;
            }
            else
            {
                // Pre-1.8.6, this always defaulted to '.'

#if defined(RECLS_PLATFORM_IS_UNIX)
                recls_char_t const* sep =   traits_type::str_pbrk(pattern, RECLS_LITERAL("|:"));
#elif defined(RECLS_PLATFORM_IS_WIN32)
                recls_char_t const* sep =   traits_type::str_pbrk(pattern, RECLS_LITERAL("|;"));
#endif /* platform */
                recls_char_t const* wc      =   NULL;

                if( NULL == sep &&
                    traits_type::is_path_rooted(pattern) &&
                    NULL != (wc = traits_type::str_pbrk(pattern, RECLS_LITERAL("*?"))))
                {
//                  const size_t    len     =   traits_type::str_len(pattern);
                    recls_char_t*   file0   =   traits_type::str_rchr(pattern, RECLS_LITERAL('/'));
#if defined(RECLS_PLATFORM_IS_UNIX)
                    recls_char_t*   file1   =   NULL;
#elif defined(RECLS_PLATFORM_IS_WIN32)
                    recls_char_t*   file1   =   traits_type::str_rchr(pattern, RECLS_LITERAL('\\'));
#endif /* platform */
                    recls_char_t*   file    =   (NULL == file0) ? file1 : (NULL == file1) ? file0 : (file0 < file1) ? file1 : file0;

                    // Not valid if wildcard comes before file, or pattern too long
                    if( wc < file ||
                        static_cast<size_t>(file - pattern) > home.size() - 1)
                    {
                        searchRoot = s_defaultSearchRoot;
                    }
                    else
                    {
                        traits_type::str_n_copy(&home[0], pattern, (file - pattern) + 1);
                        home[(file - pattern) + 1] = RECLS_LITERAL('\0');

                        searchRoot  =   home.data();
                        pattern     =   file + 1;
                    }
                }
                else
                {
                    searchRoot = s_defaultSearchRoot;
                }
            }
        }
        else
                // Handle tilde
                if(is_home_start_(searchRoot))
        {
            size_t  n   =   get_home_(&home[0], home.size());

            if(0 == n)
            {
                return RECLS_RC_NO_HOME;
            }
            else
            {
                RECLS_ASSERT(!traits_type::has_dir_end(&home[0]));
//              traits_type::ensure_dir_end(&home[n - 1]);
                traits_type::str_cat(&home[n - 1], searchRoot + 1);
                searchRoot = home.c_str();
            }
        }

        const size_t                                patternLen  =   traits_type::str_len(pattern);
        stlsoft_ns_qual(auto_buffer)<recls_char_t>  pattern_(1 + patternLen);

#ifndef _RECLS_EXCEPTION_SUPPORT
        if(0 == pattern_.size())
        {
            rc = RECLS_RC_OUT_OF_MEMORY;
        }
        else
#endif /* !_RECLS_EXCEPTION_SUPPORT */
        {
            path_type   path;

            if( NULL != traits_type::str_chr(pattern, traits_type::path_separator()) ||
                NULL != traits_type::str_chr(pattern, RECLS_LITERAL('|')))
            {
                // Has the separator, so we can proceed. (If any are too long, they will be
                // caught by IsValidPattern()

                pattern = traits_type::str_n_copy(&pattern_[0], pattern, pattern_.size());

                // We need to alter the path name separator, to the appropriate slash-form
                //
                // We also replace | with the appropriate path-name separator.

#if defined(RECLS_PLATFORM_IS_UNIX)
# if defined(WIN32) && \
     defined(EMULATE_UNIX_ON_WIN32)
                stlsoft_ns_qual_std(replace)(&pattern_[0], &pattern_[0] + patternLen, RECLS_LITERAL('\\'), RECLS_LITERAL('/'));
# endif /* WIN32 && EMULATE_UNIX_ON_WIN32 */
                stlsoft_ns_qual_std(replace)(&pattern_[0], &pattern_[0] + patternLen, RECLS_LITERAL('|'), RECLS_LITERAL(':'));

#elif defined(RECLS_PLATFORM_IS_WIN32)
                stlsoft_ns_qual_std(replace)(&pattern_[0], &pattern_[0] + patternLen, RECLS_LITERAL('/'), RECLS_LITERAL('\\'));
                stlsoft_ns_qual_std(replace)(&pattern_[0], &pattern_[0] + patternLen, RECLS_LITERAL('|'), RECLS_LITERAL(';'));
# endif /* platform */

                rc = RECLS_RC_OK;
            }
            else
            {
                // Cater for the situation whereby the pattern is a file

                if(traits_type::str_len(pattern) > traits_type::path_max())
                {
                    rc = RECLS_RC_PATH_LIMIT_EXCEEDED;
                }
                else
                {
                    rc = RECLS_RC_OK;

                    file_path_buffer_type   searchRoot_;

                    path = pattern;

                    if(path.exists())
                    {
                        path.make_absolute(true);
#ifdef _RECLS_EXCEPTION_SUPPORT
                        path.canonicalise(true);
#endif /* _RECLS_EXCEPTION_SUPPORT */

                        recls_char_t const  *file   =   path.get_file();
                        const size_t        cch     =   static_cast<size_t>(file - path.c_str());

                        traits_type::str_n_copy(&searchRoot_[0], path.c_str(), cch);
                        searchRoot_[cch] = '\0';
                        searchRoot = searchRoot_.c_str();

                        pattern = file;

                        if(0 == (flags & RECLS_F_TYPEMASK))
                        {
                            if(traits_type::is_directory(path.c_str()))
                            {
                                flags |= RECLS_F_DIRECTORIES;
                            }
                            else
                            {
                                flags |= RECLS_F_FILES;
                            }
                        }
                    }
                }
            }

            if(RECLS_SUCCEEDED(rc))
            {
                // Default the flags
                if(0 == (flags & RECLS_F_TYPEMASK))
                {
                    flags |= RECLS_F_FILES;
                }

                if(0 == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
                {
                    rc = RECLS_RC_INVALID_SEARCH_TYPE;
                }
                // Validate the pattern.
                else
                {
                    rc = IsValidPattern(pattern, flags, traits_type::path_max());
                }
            }

            if(RECLS_SUCCEEDED(rc))
            {
                ReclsFileSearch *si;

                RECLS_ASSERT(NULL != searchRoot);
                RECLS_ASSERT(NULL != pattern);

                rc = ReclsFileSearch::FindAndCreate(searchRoot, pattern, flags, pfn, param, &si);

                if(RECLS_SUCCEEDED(rc))
                {
                    *phSrch = ReclsSearch::ToHandle(si);

                    rc = RECLS_RC_OK;
                }
            }
        }
#ifdef _RECLS_EXCEPTION_SUPPORT
    }
# ifdef STLSOFT_CF_NOTHROW_BAD_ALLOC
    catch(std::bad_alloc &x)
    {
        debug_printf("Exception in Recls_Search(): %s\n", x.what());

        rc = RECLS_RC_OUT_OF_MEMORY;
    }
# endif /* STLSOFT_CF_NOTHROW_BAD_ALLOC */
    catch(std::exception &x)
    {
        debug_printf("Exception in Recls_Search(): %s\n", x.what());

        rc = RECLS_RC_UNEXPECTED;
    }
#endif /* _RECLS_EXCEPTION_SUPPORT */

    RECLS_MESSAGE_ASSERT("Must not return a success code with a null search handle", RECLS_RC_OK != rc || NULL != *phSrch);

    return rc;
}

/** Closes the given search */
RECLS_FNDECL(void) Recls_SearchClose(hrecls_t hSrch)
{
    function_scope_trace("Recls_SearchClose");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    RECLS_MESSAGE_ASSERT("Search handle is null!", NULL != si);

    delete si;
}

RECLS_FNDECL(recls_rc_t) Recls_Stat(    recls_char_t const  *path
                                    ,   recls_uint32_t      flags
                                    ,   recls_info_t        *phEntry)
{
    function_scope_trace("Recls_SearchClose");

    RECLS_ASSERT(NULL != path);
    RECLS_ASSERT(NULL != phEntry);

    *phEntry = static_cast<recls_info_t>(0);

#ifdef _RECLS_EXCEPTION_SUPPORT
    try
    {
#endif /* _RECLS_EXCEPTION_SUPPORT */
        file_path_buffer_type   home;

        if( NULL != path &&
            is_home_start_(path))
        {
            size_t  n   =   get_home_(&home[0], home.size());

            if(0 == n)
            {
                return RECLS_RC_NO_HOME;
            }
            else
            {
                RECLS_ASSERT(!traits_type::has_dir_end(&home[0]));
//              traits_type::ensure_dir_end(&home[n - 1]);
                traits_type::str_cat(&home[n - 1], path + 1);
                path = home.c_str();
            }
        }

        path_type   path_(path);

        if( !path_.exists() && 
            RECLS_F_DETAILS_LATER != (flags & (RECLS_F_DETAILS_LATER | RECLS_F_TYPEMASK))) // To allow non-existant things to be stat'd
        {
            path_.make_absolute();  // Need this, or get RECLS_RC_INVALID_DIRECTORY on a non-existant relative path

            recls_char_t const  *file   =   path_.get_file();
            path_type           path2_(path_.c_str(), static_cast<size_t>(file - path_.c_str()));

            if(!path2_.exists())
            {
                return RECLS_RC_INVALID_DIRECTORY;
            }

            return RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            if(traits_type::is_directory(path_.c_str()))
            {
                if(RECLS_F_FILES == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
                {
                    return RECLS_RC_WRONG_ENTRY_TYPE;
                }
            }
            else
            {
                if(RECLS_F_DIRECTORIES == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
                {
                    return RECLS_RC_WRONG_ENTRY_TYPE;
                }
            }

            path_.make_absolute(false);
#ifdef _RECLS_EXCEPTION_SUPPORT
            path_.canonicalise(false);
#endif /* _RECLS_EXCEPTION_SUPPORT */

            return ReclsFileSearch::Stat(path_.c_str(), flags, phEntry);
        }
#ifdef _RECLS_EXCEPTION_SUPPORT
    }
    catch(std::exception &x)
    {
        debug_printf("Exception in Recls_Stat(): %s\n", x.what());

        return RECLS_RC_UNEXPECTED;
    }
#endif /* _RECLS_EXCEPTION_SUPPORT */

#if (   !defined(RECLS_COMPILER_IS_BORLAND) || \
        __BORLANDC__ < 0x0560) && \
    (   !defined(RECLS_COMPILER_IS_MSVC) || \
        _MSC_VER < 1300)
    RECLS_MESSAGE_ASSERT("Unexpected code path", 0);

    return RECLS_RC_UNEXPECTED;
#endif /* compiler */
}

/* /////////////////////////////////////////////////////////////////////////////
 * Search enumeration
 */

// TODO: Make this a public API function in next version
static recls_rc_t Recls_SearchProcessFeedback(  recls_char_t const          *searchRoot
                                            ,   recls_char_t const          *pattern
                                            ,   recls_uint32_t              flags
                                            ,   hrecls_process_fn_t         pfn
                                            ,   recls_process_fn_param_t    param
                                            ,   hrecls_progress_fn_t        pfnProgress
                                            ,   recls_process_fn_param_t    paramProgress)
{
    function_scope_trace("Recls_SearchProcess");

    RECLS_ASSERT(NULL != pfn);

#if defined(RECLS_PLATFORM_IS_WIN32x)
    cdecl_process_fn_translator translator(pfn, param);

    if( NULL != pfn &&
        (flags & RECLS_F_CALLBACKS_STDCALL_ON_WIN32))
    {
        pfn     =   &cdecl_process_fn_translator::func;
        param   =   translator;
    }
#endif /* RECLS_PLATFORM_IS_WIN32 */

    hrecls_t    hSrch;
    recls_rc_t  rc  =   Recls_SearchFeedback(searchRoot, pattern, flags, pfnProgress, paramProgress, &hSrch);

    if(RECLS_SUCCEEDED(rc))
    {
        recls_info_t    info;

        do
        {
            rc = Recls_GetDetails(hSrch, &info);

            if(RECLS_FAILED(rc))
            {
                break;
            }
            else
            {
                int res;

#if defined(RECLS_PLATFORM_IS_WIN32)
                if(flags & RECLS_F_CALLBACKS_STDCALL_ON_WIN32)
                {
                    typedef int (RECLS_CALLCONV_STDDECL *stdcall_process_fn_t)( recls_info_t                hEntry
                                                                            ,   recls_process_fn_param_t    param);

                    stdcall_process_fn_t    pfn_stdcall =   (stdcall_process_fn_t)pfn;

                    res = (*pfn_stdcall)(info, param);
                }
                else
#endif /* RECLS_PLATFORM_IS_WIN32 */
                {
                    res = (*pfn)(info, param);
                }

                Recls_CloseDetails(info);

                if(0 == res)
                {
                    rc = RECLS_RC_SEARCH_CANCELLED;

                    break;
                }
            }
        }
        while(RECLS_SUCCEEDED(rc = Recls_GetNext(hSrch)));

        Recls_SearchClose(hSrch);
    }

    if(RECLS_RC_NO_MORE_DATA == rc)
    {
        rc = RECLS_RC_OK;
    }

    return rc;
}

RECLS_FNDECL(recls_rc_t) Recls_SearchProcess(   recls_char_t const          *searchRoot
                                            ,   recls_char_t const          *pattern
                                            ,   recls_uint32_t              flags
                                            ,   hrecls_process_fn_t         pfn
                                            ,   recls_process_fn_param_t    param)
{
    function_scope_trace("Recls_SearchProcess");

    return Recls_SearchProcessFeedback(searchRoot, pattern, flags, pfn, param, NULL, NULL);
}

RECLS_FNDECL(recls_rc_t) Recls_GetNext(hrecls_t hSrch)
{
    function_scope_trace("Recls_GetNext");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    RECLS_MESSAGE_ASSERT("Search handle is null!", NULL != si);

    return si->GetNext();
}

RECLS_FNDECL(recls_rc_t) Recls_GetDetails(  hrecls_t        hSrch
                                        ,   recls_info_t    *pinfo)
{
    function_scope_trace("Recls_GetDetails");

    ReclsSearch     *si     =   ReclsSearch::FromHandle(hSrch);

    RECLS_MESSAGE_ASSERT("Search handle is null!", NULL != si);
    RECLS_ASSERT(NULL != pinfo);

    return si->GetDetails(pinfo);
}

RECLS_FNDECL(recls_rc_t) Recls_GetNextDetails(  hrecls_t        hSrch
                                            ,   recls_info_t    *pinfo)
{
    function_scope_trace("Recls_GetNextDetails");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    RECLS_MESSAGE_ASSERT("Search handle is null!", NULL != si);

    return si->GetNextDetails(pinfo);
}

/***************************************
 * File entry info structure
 */

RECLS_FNDECL(void) Recls_CloseDetails(recls_info_t fileInfo)
{
    function_scope_trace("Recls_CloseDetails");

    RECLS_ASSERT(NULL != fileInfo);

    Entry_Release(fileInfo);
}

RECLS_FNDECL(recls_rc_t) Recls_CopyDetails( recls_info_t    fileInfo
                                        ,   recls_info_t    *pinfo)
{
    function_scope_trace("Recls_CopyDetails");

    RECLS_ASSERT(NULL != pinfo);

    return Entry_Copy(fileInfo, pinfo);
}

RECLS_FNDECL(recls_rc_t) Recls_OutstandingDetails(  hrecls_t        hSrch
                                                ,   recls_uint32_t  *count)
{
    function_scope_trace("Recls_OutstandingDetails");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    RECLS_MESSAGE_ASSERT("Search handle is null!", NULL != si);
    RECLS_ASSERT(NULL != count);
    STLSOFT_SUPPRESS_UNUSED(si);

    recls_sint32_t  cCreated;
    recls_sint32_t  cShared;

    Entry_BlockCount(&cCreated, &cShared);

    *count = static_cast<recls_uint32_t>(cCreated);

    return RECLS_RC_OK;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Error handling
 */

RECLS_FNDECL(recls_rc_t) Recls_GetLastError(hrecls_t hSrch)
{
    function_scope_trace("Recls_GetLastError");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    RECLS_MESSAGE_ASSERT("Search handle is null!", NULL != si);

    return si->GetLastError();
}

static recls_char_t const *lookup_error_string_(recls_rc_t rc, size_t *len)
{
    struct StringEntry
    {
        recls_rc_t          rc;
        recls_char_t const  *str;
        size_t              len;
    };

#define RC_STR_DECL(rc, desc)                                                       \
                                                                                    \
    static const recls_char_t   s_str##rc[] =   RECLS_LITERAL(desc);                \
    static const StringEntry    s_rct##rc = { rc, s_str##rc, NUM_ELEMENTS(desc) - 1 }


#define RC_STR_ENTRY(rc)        &s_rct##rc

#ifndef NUM_ELEMENTS
# define NUM_ELEMENTS(ar)       (sizeof(ar) / sizeof(ar[0]))
#endif /* !NUM_ELEMENTS */

    RC_STR_DECL(RECLS_RC_SEARCH_NO_CURRENT,         "Search has no current node");
    RC_STR_DECL(RECLS_RC_INVALID_DIRECTORY,         "Invalid directory");
    RC_STR_DECL(RECLS_RC_NO_MORE_DATA,              "No more data");
    RC_STR_DECL(RECLS_RC_OUT_OF_MEMORY,             "No more memory");
    RC_STR_DECL(RECLS_RC_NOT_IMPLEMENTED,           "Not implemented");
    RC_STR_DECL(RECLS_RC_INVALID_SEARCH_TYPE,       "The search type was invalid");
    RC_STR_DECL(RECLS_RC_INVALID_SEARCH_PROTOCOL,   "The search protocol was invalid");
    RC_STR_DECL(RECLS_RC_UNEXPECTED,                "An unexpected internal condition was encountered");
    RC_STR_DECL(RECLS_RC_DOT_RECURSIVE_SEARCH,      "\"..\" specifed, or \".\" specified pattern in recursive search");
    RC_STR_DECL(RECLS_RC_SEARCH_CANCELLED,          "The search was cancelled; it did not otherwise fail");
    RC_STR_DECL(RECLS_RC_FTP_INIT_FAILED,           "FTP infrastructure initialisation failed");
    RC_STR_DECL(RECLS_RC_FTP_CONNECTION_FAILED,     "FTP connection failed/rejected");
    RC_STR_DECL(RECLS_RC_WRONG_ENTRY_TYPE,          "Given entry was of wrong type");
    RC_STR_DECL(RECLS_RC_CANNOT_STAT_ROOT,          "Cannot call stat() on root");
    RC_STR_DECL(RECLS_RC_PATH_LIMIT_EXCEEDED,       "Directory or pattern component exceeds maximum for host system");
    RC_STR_DECL(RECLS_RC_USER_CANCELLED_SEARCH,     "User cancelled the search (by a 0 return from the Recls_SearchFeedback() callback function)");
    RC_STR_DECL(RECLS_RC_NO_HOME,                   "The user's home directory cannot be determined");

    static const StringEntry    *entries[] = 
    {
        RC_STR_ENTRY(RECLS_RC_SEARCH_NO_CURRENT),
        RC_STR_ENTRY(RECLS_RC_INVALID_DIRECTORY),
        RC_STR_ENTRY(RECLS_RC_NO_MORE_DATA),
        RC_STR_ENTRY(RECLS_RC_OUT_OF_MEMORY),
        RC_STR_ENTRY(RECLS_RC_NOT_IMPLEMENTED),
        RC_STR_ENTRY(RECLS_RC_INVALID_SEARCH_TYPE),
        RC_STR_ENTRY(RECLS_RC_INVALID_SEARCH_PROTOCOL),
        RC_STR_ENTRY(RECLS_RC_UNEXPECTED),
        RC_STR_ENTRY(RECLS_RC_DOT_RECURSIVE_SEARCH),
        RC_STR_ENTRY(RECLS_RC_SEARCH_CANCELLED),
        RC_STR_ENTRY(RECLS_RC_FTP_INIT_FAILED),
        RC_STR_ENTRY(RECLS_RC_FTP_CONNECTION_FAILED),
        RC_STR_ENTRY(RECLS_RC_WRONG_ENTRY_TYPE),
        RC_STR_ENTRY(RECLS_RC_CANNOT_STAT_ROOT),
        RC_STR_ENTRY(RECLS_RC_PATH_LIMIT_EXCEEDED),
        RC_STR_ENTRY(RECLS_RC_USER_CANCELLED_SEARCH),
        RC_STR_ENTRY(RECLS_RC_NO_HOME),
    };
    size_t                      len_;   // Null object pattern

    if(NULL == len)
    {
        len = &len_;
    }

    for(size_t i = 0; i < stlsoft_num_elements_(entries); ++i)
    {
        if(entries[i]->rc == rc)
        {
            return (*len = entries[i]->len, entries[i]->str);
        }
    }

    return (*len = 0, RECLS_LITERAL(""));
}

RECLS_FNDECL(size_t) Recls_GetErrorString(  recls_rc_t      rc
                                        ,   recls_char_t    *buffer
                                        ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetErrorString");

    size_t              cchError;
    recls_char_t const  *s  =   lookup_error_string_(rc, &cchError);

    if(NULL == buffer)
    {
        return cchError;
    }
    else
    {
        return recls_strncpy(buffer, cchBuffer, s, cchError);
    }
}

RECLS_FNDECL(size_t) Recls_GetLastErrorString(  hrecls_t        hSrch
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetLastErrorString");

    return Recls_GetErrorString(Recls_GetLastError(hSrch), buffer, cchBuffer);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Property elicitation
 */

RECLS_FNDECL(size_t) Recls_GetPathProperty( recls_info_t    fileInfo
                                        ,   recls_char_t    *buffer
                                        ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetPathProperty");

    RECLS_ASSERT(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->path, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetSearchRelativePathProperty(   recls_info_t    fileInfo
                                                    ,       recls_char_t    *buffer
                                                    ,       size_t          cchBuffer)
{
    function_scope_trace("Recls_GetSearchRelativePathProperty");

    RECLS_ASSERT(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->searchRelativePath, buffer, cchBuffer);
}


RECLS_FNDECL(size_t) Recls_GetDirectoryProperty(    recls_info_t    fileInfo
                                                ,   recls_char_t    *buffer
                                                ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryProperty");

    RECLS_ASSERT(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->directory, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetDirectoryPathProperty(    recls_info_t    fileInfo
                                                    ,   recls_char_t    *buffer
                                                    ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryPathProperty");

    RECLS_ASSERT(NULL != fileInfo);

    struct recls_strptrs_t directoryPath =
    {
            fileInfo->path.begin    /* Directory path is defined by start of path ... */
        ,   fileInfo->directory.end /* ... to end of directory. */
    };

    return Recls_GetStringProperty_(&directoryPath, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetSearchDirectoryProperty(  recls_info_t    fileInfo
                                                    ,   recls_char_t    *buffer
                                                    ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetSearchDirectoryProperty");

    RECLS_ASSERT(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->searchDirectory, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetUNCDriveProperty( recls_info_t    fileInfo
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetUNCDriveProperty");

    RECLS_ASSERT(NULL != fileInfo);

    struct recls_strptrs_t uncDrive =
    {
            fileInfo->path.begin        /* Directory path is defined by start of path ... */
        ,   fileInfo->directory.begin   /* ... to start of directory. */
    };

    if(!Recls_IsFileUNC(fileInfo))
    {
        uncDrive.end    =   uncDrive.begin;
    }

    return Recls_GetStringProperty_(&uncDrive, buffer, cchBuffer);
}


RECLS_FNDECL(size_t) Recls_GetFileProperty( recls_info_t    fileInfo
                                        ,   recls_char_t    *buffer
                                        ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetFileProperty");

    RECLS_ASSERT(NULL != fileInfo);

    struct recls_strptrs_t file =
    {
            fileInfo->fileName.begin /* File is defined by start of fileName ... */
        ,   fileInfo->fileExt.end    /* ... to end of fileExt. */
    };

    return Recls_GetStringProperty_(&file, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetFileNameProperty( recls_info_t    fileInfo
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetFileNameProperty");

    RECLS_ASSERT(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->fileName, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetFileExtProperty(  recls_info_t    fileInfo
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetFileExtProperty");

    RECLS_ASSERT(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->fileExt, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetDirectoryPartProperty(    recls_info_t    fileInfo
                                                    ,   int             part
                                                    ,   recls_char_t    *buffer
                                                    ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryPartProperty");

    RECLS_ASSERT(NULL != fileInfo);

    size_t  cParts = static_cast<size_t>(fileInfo->directoryParts.end - fileInfo->directoryParts.begin);

    if(part < 0)
    {
        return cParts;
    }
    else
    {
        RECLS_ASSERT(static_cast<size_t>(part) < cParts);

        return Recls_GetStringProperty_(&fileInfo->directoryParts.begin[part], buffer, cchBuffer);
    }
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileReadOnly(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileReadOnly");

    RECLS_ASSERT(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_UNIX)
    return (fileInfo->attributes & S_IWRITE) == 0;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    return fileInfo->attributes & FILE_ATTRIBUTE_READONLY;
#else /* unrecognised platform */
# error platform is not recognised
#endif /* platform */
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileDirectory(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileDirectory");

    RECLS_ASSERT(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_UNIX)
    return (fileInfo->attributes & S_IFMT) == S_IFDIR;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    return fileInfo->attributes & FILE_ATTRIBUTE_DIRECTORY;
#else /* unrecognised platform */
# error platform is not recognised
#endif /* platform */
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileLink(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileLink");

    RECLS_ASSERT(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_UNIX) && \
    (   !defined(WIN32) || \
        !defined(EMULATE_UNIX_ON_WIN32))
    return (fileInfo->attributes & S_IFMT) == S_IFLNK;
#else /* unrecognised platform */
    STLSOFT_SUPPRESS_UNUSED(fileInfo);

    return false;
#endif /* platform */
}

RECLS_FNDECL(recls_bool_t) Recls_DoesEntryExist(recls_info_t hEntry)
{
    function_scope_trace("Recls_DoesEntryExist");

    RECLS_ASSERT(NULL != hEntry);

    return traits_type::file_exists(hEntry->path.begin);
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileUNC(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetCreationTime");

    RECLS_ASSERT(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_WIN32) || \
    (   defined(RECLS_PLATFORM_IS_UNIX) && \
        defined(WIN32) && \
        defined(EMULATE_UNIX_ON_WIN32))
    return ('\\' == fileInfo->path.begin[0] && '\\' == fileInfo->path.begin[1]);
#else /* unrecognised platform */
    STLSOFT_SUPPRESS_UNUSED(fileInfo);

    return false;
#endif /* platform */
}

RECLS_FNDECL(recls_time_t) Recls_GetCreationTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetCreationTime");

    RECLS_ASSERT(NULL != fileInfo);

    return fileInfo->_GetCreationTime;
}

RECLS_FNDECL(recls_time_t) Recls_GetLastStatusChangeTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetLastStatusChangeTime");

    RECLS_ASSERT(NULL != fileInfo);

    return fileInfo->_GetLastStatusChangeTime;
}

RECLS_FNDECL(void) Recls_GetSizeProperty(   recls_info_t        fileInfo
                                        ,   recls_filesize_t    *size)
{
    function_scope_trace("Recls_GetSizeProperty");

    RECLS_ASSERT(NULL != fileInfo);
    RECLS_ASSERT(NULL != size);

    *size = fileInfo->size;
}

RECLS_FNDECL(recls_time_t) Recls_GetModificationTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetModificationTime");

    RECLS_ASSERT(NULL != fileInfo);

    return fileInfo->modificationTime;
}

RECLS_FNDECL(recls_time_t) Recls_GetLastAccessTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetLastAccessTime");

    RECLS_ASSERT(NULL != fileInfo);

    return fileInfo->lastAccessTime;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Roots
 */

RECLS_LINKAGE_C recls_char_t const *Recls_GetPathNameSeparator(void)
{
    static recls_char_t const   s_pathNameSeparator[] = { RAPI_PATHNAMESEP, '\0' };

    return s_pathNameSeparator;
}

RECLS_LINKAGE_C recls_char_t const *Recls_GetPathSeparator(void)
{
    static recls_char_t const   s_pathSeparator[] = { RAPI_PATHSEP, '\0' };

    return s_pathSeparator;
}

RECLS_LINKAGE_C recls_char_t const *Recls_GetWildcardsAll(void)
{
    static recls_char_t const   s_wildcardsAll[] = { RAPI_WILDCARDSALL, '\0' };

    return s_wildcardsAll;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Extended API functions
 */

static int RECLS_CALLCONV_DEFAULT IsDirectoryEmpty_proc(recls_info_t /* info */, recls_process_fn_param_t /* param */)
{
    return 0; // Cancel on any entry
}

RECLS_FNDECL(recls_bool_t) Recls_IsDirectoryEmpty(recls_char_t const *dir)
{
    recls_rc_t  rc  =   Recls_SearchProcess(dir, Recls_GetWildcardsAll(), RECLS_F_TYPEMASK, IsDirectoryEmpty_proc, 0);

    return RECLS_RC_SEARCH_CANCELLED != rc;
}

RECLS_FNDECL(recls_bool_t) Recls_IsDirectoryEntryEmpty(recls_info_t hEntry)
{
    RECLS_ASSERT(NULL != hEntry);
    RECLS_ASSERT(Recls_IsFileDirectory(hEntry));

    return Recls_IsDirectoryEmpty(hEntry->path.begin);
}

static int RECLS_CALLCONV_DEFAULT CalcDirectorySize_proc(recls_info_t hEntry, recls_process_fn_param_t param)
{
    recls_filesize_t    &total = *static_cast<recls_filesize_t*>(param);
    recls_filesize_t    size;

    Recls_GetSizeProperty(hEntry, &size);

#if defined(RECLS_PLATFORM_IS_UNIX)
    total += size;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    total.QuadPart += size.QuadPart;
#else /* ? platform */
# error Platform not (yet) discriminated
#endif /* platform */

    return 1; // Never cancel
}

RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectorySize(recls_char_t const *dir)
{
    static recls_filesize_t zero;
    recls_filesize_t        total = zero;

    Recls_SearchProcess(dir, Recls_GetWildcardsAll(), RECLS_F_FILES | RECLS_F_RECURSIVE, CalcDirectorySize_proc, static_cast<recls_process_fn_param_t>(&total));

    return total;
}

RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectoryEntrySize(recls_info_t hEntry)
{
    RECLS_ASSERT(NULL != hEntry);
    RECLS_ASSERT(Recls_IsFileDirectory(hEntry));

    return Recls_CalcDirectorySize(hEntry->path.begin);
}



RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectorySizeFeedback( recls_char_t const          *dir
                                                            ,   recls_uint32_t              flags
                                                            ,   hrecls_progress_fn_t        pfn
                                                            ,   recls_process_fn_param_t    param)
{
    static recls_filesize_t zero;
    recls_filesize_t        total = zero;

    flags   &=  ~(RECLS_F_DIRECTORIES);
    flags   |=  RECLS_F_FILES;
    flags   |=  RECLS_F_RECURSIVE;

    Recls_SearchProcessFeedback(dir, Recls_GetWildcardsAll(), flags, CalcDirectorySize_proc, static_cast<recls_process_fn_param_t>(&total), pfn, param);

    return total;
}

RECLS_FNDECL(size_t) Recls_CombinePaths(recls_char_t const  *path1
                                    ,   recls_char_t const  *path2
                                    ,   recls_char_t        *result
                                    ,   size_t              cchResult)
{
    RECLS_ASSERT(NULL != path1 || NULL != path2);

    platformstl::basic_path<recls_char_t>   path((NULL != path1) ? path1 : path2);

    if( NULL != path1 &&
        NULL != path2)
    {
        path /= path2;
    }

    return path.copy(result, cchResult);
}


RECLS_FNDECL(size_t) Recls_DeriveRelativePath(  recls_char_t const  *origin
                                            ,   recls_char_t const  *target
                                            ,   recls_char_t        *result
                                            ,   size_t              cchResult)
{
    typedef platformstl::basic_path<recls_char_t>           path_t;
    typedef platformstl::filesystem_traits<recls_char_t>    traits_t;

    path_t      originPath(origin);
    path_t      targetPath(target);
    const bool  bTargetHasSeparator =   targetPath.has_sep();

    // Make paths absolute, canonicalise, and remove any trailing separators
    originPath.make_absolute().canonicalise().pop_sep();
    targetPath.make_absolute().canonicalise().pop_sep();

#ifdef _WIN32
    if( traits_t::is_path_UNC(originPath.c_str()) ||
        traits_t::is_path_UNC(targetPath.c_str()))
    {
        RECLS_ASSERT('\\' == originPath[0] && '\\' == originPath[1]);
        RECLS_ASSERT('\\' == targetPath[0] && '\\' == targetPath[1]);

        recls_char_t const  *originShare    =   traits_t::str_chr(&originPath[2], '\\');
        recls_char_t const  *targetShare    =   traits_t::str_chr(&targetPath[2], '\\');

        if( (originShare - &originPath[2]) != (targetShare - &targetPath[2]) ||
            0 != traits_t::str_n_compare(&originPath[2], &targetPath[2], static_cast<size_t>(originShare - &originPath[2])))
        {
            // Different shares, so return target

            return targetPath.copy(result, cchResult);
        }
    }
    else
    {
        RECLS_ASSERT(isalpha(originPath[0]) && ':' == originPath[1]);
        RECLS_ASSERT(isalpha(targetPath[0]) && ':' == targetPath[1]);

        if(toupper(originPath[0]) != toupper(targetPath[0]))
        {
            // Different drives, so return target

            return targetPath.copy(result, cchResult);
        }
    }
#endif /* _WIN32 */

    // now trim the common 
    recls_char_t const* po  =   &originPath[0];
    recls_char_t const* pt  =   &targetPath[0];

    for(;;)
    {
#ifdef _WIN32
        if(toupper(*po) != toupper(*pt))
#else /* ? _WIN32 */
        if(*po != *pt)
#endif /* _WIN32 */
        {
            break;
        }

        ++po;
        ++pt;

        if('\0' == *po)
        {
            break;
        }
        if('\0' == *pt)
        {
            break;
        }
    }

    // We may need to rewind, since "h:\abc\def" and "H:\abc\defghi" would yield
    // ghi

    if( po != &originPath[0] &&
        (   '\0' != *po ||
            '\0' != *pt))
    {
        // There is some commonality

        if(traits_t::is_path_name_separator(*(po - 1)))
        {
            // Don't do anything, because both strings are at a unique whole point
        }
        else
        {
            // Previous was not a separator

            if( (   '\0' == *po &&
                    traits_t::is_path_name_separator(*pt)) ||
                (   '\0' == *pt &&
                    traits_t::is_path_name_separator(*po)))
            {
                // One is at end, other at path-name-sep
            }
            else
            {
                for(; po != &originPath[0] && !traits_t::is_path_name_separator(*(po - 1)); --po)
                {}
                for(; pt != &originPath[0] && !traits_t::is_path_name_separator(*(pt - 1)); --pt)
                {}
            }
        }

        if('\0' == *po)
        {
            if(traits_t::is_path_name_separator(*pt))
            {
                ++pt;
            }
        }
        else if('\0' == *pt)
        {
            if(traits_t::is_path_name_separator(*po))
            {
                ++po;
            }
        }

    }

    // Now we take any dir parts from target, and turn them into .. for the
    // origin.

    path_t  originFragment(po);
    path_t  targetFinal;

    for(; 0 != originFragment.size(); originFragment.pop(true))
    {
        targetFinal /= RECLS_LITERAL("..");
    }

    targetFinal /= pt;

    if(bTargetHasSeparator)
    {
        targetFinal.push_sep();
    }

    return targetFinal.copy(result, cchResult);
}

RECLS_FNDECL(size_t) Recls_SqueezePath( recls_char_t const  *path
                                    ,   recls_char_t        *result
                                    ,   size_t              cchResult)
{
    return platformstl::path_squeeze(path, result, cchResult);
}

/* /////////////////////////////////////////////////////////////////////////
 * Unit-tests
 */

RECLS_FNDECL(void) Recls_RunUnittests_EntryInfo(FILE *stm);

RECLS_FNDECL(void) Recls_RunUnittests(FILE *stm)
{
    Recls_RunUnittests_EntryInfo(stm);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
