/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_util_win32.cpp
 *
 * Purpose:     Win32 utility functions for the recls API.
 *
 * Created:     17th August 2003
 * Updated:     14th June 2006
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2006, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/recls.h>
#include <recls/assert.h>
#include "recls_util.h"
#include "recls_impl.h"
#include "recls_debug.h"

#include <winstl/filesystem/filesystem_traits.hpp>

#if RECLS_DEBUG_LEVEL > 0
# include <ctype.h>
#endif /* RECLS_DEBUG_LEVEL */

#include "recls_debug.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

typedef winstl_ns_qual(filesystem_traits)<recls_char_t> traits_t;

/* ////////////////////////////////////////////////////////////////////////// */

RECLS_LINKAGE_C recls_char_t const *find_directory_0(recls_char_t const *path)
{
    if(':' == path[1])
    {
        // It's a drive-prefixed absolute path, so ...
#if RECLS_DEBUG_LEVEL > 0
        if(!isalpha(path[0]))
        {
            debug_printf("find_directory_0() given an invalid path: %s", path);
        }
#endif /* RECLS_DEBUG_LEVEL > 0 */

        // ... we just skip the drive
        return &path[2];
    }
    else if('\\' == path[0] &&
            '\\' == path[1])
    {
        // It's a UNC absolute path, so we have to find the share name (with a '\')
        // and then the next slash or backslash
        recls_char_t const  *share  =   traits_t::str_chr(path + 2, '\\');

        if(NULL == share)
        {
            goto bad_path_given;
        }
        else
        {
            recls_char_t const  *slash      =   traits_t::str_chr(share + 1, '\\');
            recls_char_t const  *slash_a    =   traits_t::str_chr(share + 1, '/');

            if( NULL == slash ||
                (   NULL != slash_a &&
                    slash_a < slash))
            {
                slash = slash_a;
            }

            if(NULL == slash)
            {
                goto bad_path_given;
            }
            else
            {
                return slash;
            }
        }
    }
    else
    {
        RECLS_ASSERT(2 < traits_t::str_len(path));

        return path;
    }

bad_path_given:

    // Can't really do _anything_ sensible here, so we just return the
    // end of the string.
#if RECLS_DEBUG_LEVEL > 0
    debug_printf("find_directory_0() given an invalid path: %s", path);
#endif /* RECLS_DEBUG_LEVEL > 0 */

    return path + traits_t::str_len(path);
}

RECLS_LINKAGE_C size_t get_home_(recls_char_t *buff, size_t cchBuff)
{
    recls_char_t    homeDrive[1 + _MAX_DRIVE];
    recls_char_t    homeDir[1 + _MAX_DIR];

    const size_t    cchHomeDrive    =   traits_t::get_environment_variable( RECLS_LITERAL("HOMEDRIVE")
                                                                        ,   &homeDrive[0]
                                                                        ,   RECLS_NUM_ELEMENTS(homeDrive));
    const size_t    cchHomeDir      =   traits_t::get_environment_variable( RECLS_LITERAL("HOMEPATH")
                                                                        ,   &homeDir[0]
                                                                        ,   RECLS_NUM_ELEMENTS(homeDir));

    if( 0 == cchHomeDrive ||
        RECLS_NUM_ELEMENTS(homeDrive) == cchHomeDrive)
    {
        return 0;
    }
    if( 0 == cchHomeDir ||
        RECLS_NUM_ELEMENTS(homeDir) == cchHomeDir)
    {
        return 0;
    }

    if(NULL == buff)
    {
        return cchHomeDrive + cchHomeDir;
    }
    else
    {
        if(cchBuff <= cchHomeDrive)
        {
            traits_t::str_n_copy(buff, homeDrive, cchBuff);

            return cchHomeDrive;
        }
        else if(cchBuff <= cchHomeDrive + cchHomeDir)
        {
            traits_t::str_copy(buff, homeDrive);
            traits_t::str_n_copy(&buff[cchHomeDrive], homeDir, cchBuff - cchHomeDrive);

            return cchBuff;
        }
        else
        {
            traits_t::str_copy(buff, homeDrive);
            traits_t::str_n_copy(&buff[cchHomeDrive], homeDir, cchHomeDir);
            buff[cchHomeDrive + cchHomeDir] = '\0';

            return cchHomeDrive + cchHomeDir;
        }
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
